# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A module for drawing circuits in ascii art or some other text representation
"""

import sys
from shutil import get_terminal_size
from warnings import warn
import numpy as np

from QuICT.core.gate import *


MAX_FRAC = 16
N, D = np.meshgrid(np.arange(1, MAX_FRAC + 1), np.arange(1, MAX_FRAC + 1))
FRAC_MESH = N / D * np.pi
RECIP_MESH = N / D / np.pi
POW_LIST = np.pi ** np.arange(2, 5)


class circuit_layer(object):
    def __init__(self):
        self.pic = set()
        self.occupy = set()
        self.gates = []

    def addGate(self, gate: BasicGate) -> bool:
        Q_set = set(gate.cargs) | set(gate.targs)
        for element in range(min(Q_set), max(Q_set) + 1):
            if element in self.pic:
                return False
        for element in range(min(Q_set), max(Q_set) + 1):
            self.pic.add(element)
        self.occupy |= Q_set
        self.gates.append(gate)
        return True

    def checkGate(self, gate: BasicGate) -> bool:
        Q_set = set(gate.cargs) | set(gate.targs)
        for element in range(min(Q_set), max(Q_set) + 1):
            if element in self.pic:
                return False
        return True


def pi_check(inpt, eps=1e-6, output='text', ndigits=5):
    """ Computes if a number is close to an integer
    fraction or multiple of PI and returns the
    corresponding string.

    Args:
        inpt (float): Number to check.
        eps (float): EPS to check against.
        output (str): Options are 'text' (default),
                      'latex', 'mpl', and 'qasm'.
        ndigits (int): Number of digits to print
                       if returning raw inpt.

    Returns:
        str: string representation of output.

    """

    def normalize(single_inpt):
        if abs(single_inpt) < 1e-14:
            return '0'

        if output == 'text':
            pi = 'π'
        elif output == 'qasm':
            pi = 'pi'
        elif output == 'latex':
            pi = '\\pi'
        elif output == 'mpl':
            pi = '$\\pi$'
        else:
            raise Exception('pi_check parameter output should be text, '
                            'latex, mpl, or qasm.')

        neg_str = '-' if single_inpt < 0 else ''

        # First check is for whole multiples of pi
        val = single_inpt / np.pi
        if abs(val) >= 1 - eps:
            if abs(abs(val) - abs(round(val))) < eps:
                val = int(abs(round(val)))
                if abs(val) == 1:
                    str_out = '{}{}'.format(neg_str, pi)
                else:
                    if output == 'qasm':
                        str_out = '{}{}*{}'.format(neg_str, val, pi)
                    else:
                        str_out = '{}{}{}'.format(neg_str, val, pi)
                return str_out

        # Second is a check for powers of pi
        if abs(single_inpt) > np.pi:
            power = np.where(abs(abs(single_inpt) - POW_LIST) < eps)
            if power[0].shape[0]:
                if output == 'qasm':
                    str_out = '{:.{}g}'.format(single_inpt, ndigits)
                elif output == 'latex':
                    str_out = '{}{}^{}'.format(neg_str, pi, power[0][0] + 2)
                elif output == 'mpl':
                    str_out = '{}{}$^{}$'.format(neg_str, pi, power[0][0] + 2)
                else:
                    str_out = '{}{}**{}'.format(neg_str, pi, power[0][0] + 2)
                return str_out

        # Third is a check for a number larger than MAX_FRAC * pi, not a
        # multiple or power of pi, since no fractions will exceed MAX_FRAC * pi
        if abs(single_inpt) >= (MAX_FRAC * np.pi):
            str_out = '{:.{}g}'.format(single_inpt, ndigits)
            return str_out

        # Fourth check is for fractions for 1*pi in the numer and any
        # number in the denom.
        val = np.pi / single_inpt
        if abs(abs(val) - abs(round(val))) < eps:
            val = int(abs(round(val)))
            if output == 'latex':
                str_out = '\\frac{%s%s}{%s}' % (neg_str, pi, val)
            else:
                str_out = '{}{}/{}'.format(neg_str, pi, val)
            return str_out

        # Fifth check is for fractions where the numer > 1*pi and numer
        # is up to MAX_FRAC*pi and denom is up to MAX_FRAC and all
        # fractions are reduced. Ex. 15pi/16, 2pi/5, 15pi/2, 16pi/9.
        frac = np.where(np.abs(abs(single_inpt) - FRAC_MESH) < eps)
        if frac[0].shape[0]:
            numer = int(frac[1][0]) + 1
            denom = int(frac[0][0]) + 1
            if output == 'latex':
                str_out = '\\frac{%s%s%s}{%s}' % (neg_str, numer, pi, denom)
            elif output == 'qasm':
                str_out = '{}{}*{}/{}'.format(neg_str, numer, pi, denom)
            else:
                str_out = '{}{}{}/{}'.format(neg_str, numer, pi, denom)
            return str_out

        # Sixth check is for fractions where the numer > 1 and numer
        # is up to MAX_FRAC and denom is up to MAX_FRAC*pi and all
        # fractions are reduced. Ex. 15/16pi, 2/5pi, 15/2pi, 16/9pi
        frac = np.where(np.abs(abs(single_inpt) - RECIP_MESH) < eps)
        if frac[0].shape[0]:
            numer = int(frac[1][0]) + 1
            denom = int(frac[0][0]) + 1
            if denom == 1 and output != 'qasm':
                denom = ''
            if output == 'latex':
                str_out = '\\frac{%s%s}{%s%s}' % (neg_str, numer, denom, pi)
            elif output == 'qasm':
                str_out = '{}{}/({}*{})'.format(neg_str, numer, denom, pi)
            else:
                str_out = '{}{}/{}{}'.format(neg_str, numer, denom, pi)
            return str_out

        # Nothing found
        str_out = '{:.{}g}'.format(single_inpt, ndigits)
        return str_out

    complex_inpt = complex(inpt)
    real, imag = map(normalize, [complex_inpt.real, complex_inpt.imag])

    jstr = '\\jmath' if output == 'latex' else 'j'
    if real == '0' and imag != '0':
        str_out = imag + jstr
    elif real != '0' and imag != '0':
        op_str = '+'
        # Remove + if imag negative except for latex fractions
        if complex_inpt.imag < 0 and (output != 'latex' or '\\frac' not in imag):
            op_str = ''
        str_out = '{}{}{}{}'.format(real, op_str, imag, jstr)
    else:
        str_out = real
    return str_out


class DrawElement:
    """ An element is an instruction or an operation that need to be drawn."""

    def __init__(self, label=None):
        self._width = None
        self.label = self.mid_content = label
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = self.bot_connect = " "
        self.top_pad = self._mid_padding = self.bot_pad = " "
        self.mid_bck = self.top_bck = self.bot_bck = " "
        self.bot_connector = {}
        self.top_connector = {}
        self.right_fill = self.left_fill = self.layer_width = 0
        self.wire_label = ""

    @property
    def top(self):
        """ Constructs the top line of the element"""
        if (self.width % 2) == 0 and len(self.top_format) % 2 == 1 and len(self.top_connect) == 1:
            ret = self.top_format % (self.top_pad + self.top_connect).center(self.width, self.top_pad)
        else:
            ret = self.top_format % self.top_connect.center(self.width, self.top_pad)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self.top_pad)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self.top_pad)
        ret = ret.center(self.layer_width, self.top_bck)
        return ret

    @property
    def mid(self):
        """ Constructs the middle line of the element"""
        ret = self.mid_format % self.mid_content.center(
            self.width, self._mid_padding)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self._mid_padding)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self._mid_padding)
        ret = ret.center(self.layer_width, self.mid_bck)
        return ret

    @property
    def bot(self):
        """ Constructs the bottom line of the element"""
        if (self.width % 2) == 0 and len(self.top_format) % 2 == 1:
            ret = self.bot_format % (self.bot_pad + self.bot_connect).center(self.width, self.bot_pad)
        else:
            ret = self.bot_format % self.bot_connect.center(self.width, self.bot_pad)
        if self.right_fill:
            ret = ret.ljust(self.right_fill, self.bot_pad)
        if self.left_fill:
            ret = ret.rjust(self.left_fill, self.bot_pad)
        ret = ret.center(self.layer_width, self.bot_bck)
        return ret

    @property
    def length(self):
        """ Returns the length of the element, including the box around."""
        return max(len(self.top), len(self.mid), len(self.bot))

    @property
    def width(self):
        """ Returns the width of the label, including padding"""
        if self._width:
            return self._width
        return len(self.mid_content)

    @width.setter
    def width(self, value):
        self._width = value

    def connect(self, wire_char, where, label=None):
        """Connects boxes and elements using wire_char and setting proper connectors.

        Args:
            wire_char (char): For example '║' or '│'.
            where (list["top", "bot"]): Where the connector should be set.
            label (string): Some connectors have a label (see cu1, for example).
        """

        if 'top' in where and self.top_connector:
            self.top_connect = self.top_connector[wire_char]

        if 'bot' in where and self.bot_connector:
            self.bot_connect = self.bot_connector[wire_char]

        if label:
            self.top_format = self.top_format[:-1] + (label if label else "")


class BoxOnClWire(DrawElement):
    """Draws a box on the classical wire.

    ::

        top: ┌───┐   ┌───┐
        mid: ╡ A ╞ ══╡ A ╞══
        bot: └───┘   └───┘
    """

    def __init__(self, label="", top_connect='─', bot_connect='─'):
        super().__init__(label)
        self.top_format = "┌─%s─┐"
        self.mid_format = "╡ %s ╞"
        self.bot_format = "└─%s─┘"
        self.top_pad = self.bot_pad = '─'
        self.mid_bck = '═'
        self.top_connect = top_connect
        self.bot_connect = bot_connect
        self.mid_content = label


class BoxOnQuWire(DrawElement):
    """Draws a box on the quantum wire.

    ::

        top: ┌───┐   ┌───┐
        mid: ┤ A ├ ──┤ A ├──
        bot: └───┘   └───┘
    """

    def __init__(self, label="", top_connect='─', conditional=False):
        super().__init__(label)
        self.top_format = "┌─%s─┐"
        self.mid_format = "┤ %s ├"
        self.bot_format = "└─%s─┘"
        self.top_pad = self.bot_pad = self.mid_bck = '─'
        self.top_connect = top_connect
        self.bot_connect = '╥' if conditional else '─'
        self.mid_content = label
        self.top_connector = {"│": '┴'}
        self.bot_connector = {"│": '┬'}


class MeasureTo(DrawElement):
    """The element on the classic wire to which the measure is performed.

    ::

        top:  ║     ║
        mid: ═╩═ ═══╩═══
        bot:
    """

    def __init__(self, label=''):
        super().__init__()
        self.top_connect = " ║ "
        self.mid_content = "═╩═"
        self.bot_connect = label
        self.mid_bck = "═"


class MeasureFrom(BoxOnQuWire):
    """The element on the quantum wire in which the measure is performed.

    ::

        top: ┌─┐    ┌─┐
        mid: ┤M├ ───┤M├───
        bot: └─┘    └─┘
    """

    def __init__(self):
        super().__init__()
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = "┌─┐"
        self.mid_content = "┤M├"
        self.bot_connect = "└─┘"

        self.top_pad = self.bot_pad = " "
        self._mid_padding = '─'


class MultiBox(DrawElement):
    """Elements that is draw on over multiple wires."""

    def center_label(self, input_length, order):
        """In multi-bit elements, the label is centered vertically.

        Args:
            input_length (int): Rhe amount of wires affected.
            order (int): Which middle element is this one?
        """
        if input_length == order == 0:
            self.top_connect = self.label
            return
        location_in_the_box = '*'.center(input_length * 2 - 1).index('*') + 1
        top_limit = order * 2 + 2
        bot_limit = top_limit + 2
        if top_limit <= location_in_the_box < bot_limit:
            if location_in_the_box == top_limit:
                self.top_connect = self.label
            elif location_in_the_box == top_limit + 1:
                self.mid_content = self.label
            else:
                self.bot_connect = self.label

    @property
    def width(self):
        """ Returns the width of the label, including padding"""
        if self._width:
            return self._width
        return len(self.label)


class BoxOnQuWireTop(MultiBox, BoxOnQuWire):
    """ Draws the top part of a box that affects more than one quantum wire"""

    def __init__(self, label="", top_connect=None, wire_label=''):
        super().__init__(label)
        self.wire_label = wire_label
        self.bot_connect = self.bot_pad = " "
        self.mid_content = ""  # The label will be put by some other part of the box.
        self.left_fill = len(self.wire_label)
        self.top_format = "┌─" + "s".center(self.left_fill + 1, '─') + "─┐"
        self.top_format = self.top_format.replace('s', '%s')
        self.mid_format = "┤{} %s ├".format(self.wire_label)
        self.bot_format = "│{} %s │".format(self.bot_pad * self.left_fill)
        self.top_connect = top_connect if top_connect else '─'


class BoxOnWireMid(MultiBox):
    """ A generic middle box"""

    def __init__(self, label, input_length, order, wire_label=''):
        super().__init__(label)
        self.top_pad = self.bot_pad = self.top_connect = self.bot_connect = " "
        self.wire_label = wire_label
        self.left_fill = len(self.wire_label)
        self.top_format = "│{} %s │".format(self.top_pad * self.left_fill)
        self.bot_format = "│{} %s │".format(self.bot_pad * self.left_fill)
        self.top_connect = self.bot_connect = self.mid_content = ''
        self.center_label(input_length, order)


class BoxOnQuWireMid(BoxOnWireMid, BoxOnQuWire):
    """ Draws the middle part of a box that affects more than one quantum wire"""

    def __init__(self, label, input_length, order, wire_label='', control_label=None):
        super().__init__(label, input_length, order, wire_label=wire_label)
        if control_label:
            self.mid_format = "{}{} %s ├".format(control_label, self.wire_label)
        else:
            self.mid_format = "┤{} %s ├".format(self.wire_label)


class BoxOnQuWireBot(MultiBox, BoxOnQuWire):
    """ Draws the bottom part of a box that affects more than one quantum wire"""

    def __init__(self, label, input_length, bot_connect=None, wire_label='', conditional=False):
        super().__init__(label)
        self.wire_label = wire_label
        self.top_pad = " "
        self.left_fill = len(self.wire_label)
        self.top_format = "│{} %s │".format(self.top_pad * self.left_fill)
        self.mid_format = "┤{} %s ├".format(self.wire_label)
        self.bot_format = "└─" + "s".center(self.left_fill + 1, '─') + "─┘"
        self.bot_format = self.bot_format.replace('s', '%s')
        bot_connect = bot_connect if bot_connect else '─'
        self.bot_connect = '╥' if conditional else bot_connect

        self.mid_content = self.top_connect = ""
        if input_length <= 2:
            self.top_connect = label


class BoxOnClWireTop(MultiBox, BoxOnClWire):
    """ Draws the top part of a conditional box that affects more than one classical wire"""

    def __init__(self, label="", top_connect=None, wire_label=''):
        super().__init__(label)
        self.wire_label = wire_label
        self.mid_content = ""  # The label will be put by some other part of the box.
        self.bot_format = "│ %s │"
        self.top_connect = top_connect if top_connect else '─'
        self.bot_connect = self.bot_pad = " "


class BoxOnClWireMid(BoxOnWireMid, BoxOnClWire):
    """ Draws the middle part of a conditional box that affects more than one classical wire"""

    def __init__(self, label, input_length, order, wire_label='', **_):
        super().__init__(label, input_length, order, wire_label=wire_label)
        self.mid_format = "╡{} %s ╞".format(self.wire_label)


class BoxOnClWireBot(MultiBox, BoxOnClWire):
    """ Draws the bottom part of a conditional box that affects more than one classical wire"""

    def __init__(self, label, input_length, bot_connect='─', wire_label='', **_):
        super().__init__(label)
        self.wire_label = wire_label
        self.left_fill = len(self.wire_label)
        self.top_pad = ' '
        self.bot_pad = '─'
        self.top_format = "│{} %s │".format(self.top_pad * self.left_fill)
        self.mid_format = "╡{} %s ╞".format(self.wire_label)
        self.bot_format = "└─" + "s".center(self.left_fill + 1, '─') + "─┘"
        self.bot_format = self.bot_format.replace('s', '%s')
        bot_connect = bot_connect if bot_connect else '─'
        self.bot_connect = bot_connect

        self.mid_content = self.top_connect = ""
        if input_length <= 2:
            self.top_connect = label


class DirectOnQuWire(DrawElement):
    """
    Element to the wire (without the box).
    """

    def __init__(self, label=""):
        super().__init__(label)
        self.top_format = ' %s '
        self.mid_format = '─%s─'
        self.bot_format = ' %s '
        self._mid_padding = self.mid_bck = '─'
        self.top_connector = {"│": '│'}
        self.bot_connector = {"│": '│'}


class Barrier(DirectOnQuWire):
    """Draws a barrier.

    ::

        top:  ░     ░
        mid: ─░─ ───░───
        bot:  ░     ░
    """

    def __init__(self, label=""):
        super().__init__("░")
        self.top_connect = "░"
        self.bot_connect = "░"
        self.top_connector = {}
        self.bot_connector = {}


class Ex(DirectOnQuWire):
    """Draws an X (usually with a connector). E.g. the top part of a swap gate.

    ::

        top:
        mid: ─X─ ───X───
        bot:  │     │
    """

    def __init__(self, bot_connect=" ", top_connect=" ", conditional=False):
        super().__init__("X")
        self.bot_connect = "║" if conditional else bot_connect
        self.top_connect = top_connect


class Reset(DirectOnQuWire):
    """ Draws a reset gate"""

    def __init__(self, conditional=False):
        super().__init__("|0>")
        if conditional:
            self.bot_connect = "║"


class Bullet(DirectOnQuWire):
    """ Draws a bullet (usually with a connector). E.g. the top part of a CX gate.

    ::

        top:
        mid: ─■─  ───■───
        bot:  │      │
    """

    def __init__(self, top_connect="", bot_connect="", conditional=False,
                 label=None, bottom=False):
        super().__init__('■')
        self.top_connect = top_connect
        self.bot_connect = '║' if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = '─'


class OpenBullet(DirectOnQuWire):
    """Draws an open bullet (usually with a connector). E.g. the top part of a CX gate.

    ::

        top:
        mid: ─o─  ───o───
        bot:  │      │
    """

    def __init__(self, top_connect="", bot_connect="", conditional=False,
                 label=None, bottom=False):
        super().__init__('o')
        self.top_connect = top_connect
        self.bot_connect = '║' if conditional else bot_connect
        if label and bottom:
            self.bot_connect = label
        elif label:
            self.top_connect = label
        self.mid_bck = '─'


class EmptyWire(DrawElement):
    """This element is just the wire, with no instructions nor operations."""

    def __init__(self, wire):
        super().__init__(wire)
        self._mid_padding = self.mid_bck = wire

    @staticmethod
    def fillup_layer(layer, first_clbit):
        """Given a layer, replace the Nones in it with EmptyWire elements.

        Args:
            layer (list): The layer that contains Nones.
            first_clbit (int): The first wire that is classic.

        Returns:
            list: The new layer, with no Nones.
        """
        for nones in [i for i, x in enumerate(layer) if x is None]:
            layer[nones] = EmptyWire('═') if nones >= first_clbit else EmptyWire('─')
        return layer


class BreakWire(DrawElement):
    """ This element is used to break the drawing in several pages."""

    def __init__(self, arrow_char):
        super().__init__()
        self.top_format = self.mid_format = self.bot_format = "%s"
        self.top_connect = arrow_char
        self.mid_content = arrow_char
        self.bot_connect = arrow_char

    @staticmethod
    def fillup_layer(layer_length, arrow_char):
        """Creates a layer with BreakWire elements.

        Args:
            layer_length (int): The length of the layer to create
            arrow_char (char): The char used to create the BreakWire element.

        Returns:
            list: The new layer.
        """
        breakwire_layer = []
        for _ in range(layer_length):
            breakwire_layer.append(BreakWire(arrow_char))
        return breakwire_layer


class InputWire(DrawElement):
    """ This element is the label and the initial value of a wire."""

    def __init__(self, label):
        super().__init__(label)

    @staticmethod
    def fillup_layer(names):
        """Creates a layer with InputWire elements.

        Args:
            names (list): List of names for the wires.

        Returns:
            list: The new layer
        """
        longest = max([len(name) for name in names])
        inputs_wires = []
        for name in names:
            inputs_wires.append(InputWire(name.rjust(longest)))
        return inputs_wires


class TextDrawing:
    """ The text drawing"""

    def __init__(self, qregs, gates,
                 line_length=None, vertical_compression='high', initial_state=True,
                 encoding=None):
        self.qregs = qregs
        self.layers = self.resolution_layers(gates)
        self.layout = None
        self.initial_state = initial_state
        self.plotbarriers = True
        self.line_length = line_length
        if vertical_compression not in ['high', 'medium', 'low']:
            raise ValueError("Vertical compression can only be 'high', 'medium', or 'low'")
        self.vertical_compression = vertical_compression
        if encoding:
            self.encoding = encoding
        else:
            if sys.stdout.encoding:
                self.encoding = sys.stdout.encoding
            else:
                self.encoding = 'utf8'

    def __str__(self):
        return self.single_string()

    def _repr_html_(self):
        return '<pre style="word-wrap: normal;' \
               'white-space: pre;' \
               'background: #fff0;' \
               'line-height: 1.1;' \
               'font-family: &quot;Courier New&quot;,Courier,monospace">' \
               '%s</pre>' % self.single_string()

    def __repr__(self):
        return self.single_string()

    def single_string(self):
        """Creates a long string with the ascii art.
        Returns:
            str: The lines joined by a newline (``\\n``)
        """
        try:
            return "\n".join(self.lines()).encode(self.encoding).decode(self.encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            warn('The encoding %s has a limited charset. Consider a different encoding in your '
                 'environment. UTF-8 is being used instead' % self.encoding, RuntimeWarning)
            self.encoding = 'utf-8'
            return "\n".join(self.lines()).encode(self.encoding).decode(self.encoding)

    def dump(self, filename, encoding=None):
        """Dumps the ascii art in the file.

        Args:
            filename (str): File to dump the ascii art.
            encoding (str): Optional. Force encoding, instead of self.encoding.
        """
        with open(filename, mode='w', encoding=encoding or self.encoding) as text_file:
            text_file.write(self.single_string())

    def lines(self, line_length=None):
        """Generates a list with lines. These lines form the text drawing.

        Args:
            line_length (int): Optional. Breaks the circuit drawing to this length. This
                               useful when the drawing does not fit in the console. If
                               None (default), it will try to guess the console width using
                               shutil.get_terminal_size(). If you don't want pagination
                               at all, set line_length=-1.

        Returns:
            list: A list of lines with the text drawing.
        """
        if line_length is None:
            line_length = self.line_length
        if not line_length:
            if ('ipykernel' in sys.modules) and ('spyder' not in sys.modules):
                line_length = 80
            else:
                line_length, _ = get_terminal_size()

        noqubits = len(self.qregs)

        layers = self.build_layers()

        layer_groups = [[]]
        rest_of_the_line = line_length
        for layerno, layer in enumerate(layers):
            # Replace the Nones with EmptyWire
            layers[layerno] = EmptyWire.fillup_layer(layer, noqubits)

            TextDrawing.normalize_width(layer)

            if line_length == -1:
                # Do not use pagination (aka line breaking. aka ignore line_length).
                layer_groups[-1].append(layer)
                continue

            # chop the layer to the line_length (pager)
            layer_length = layers[layerno][0].length

            if layer_length < rest_of_the_line:
                layer_groups[-1].append(layer)
                rest_of_the_line -= layer_length
            else:
                layer_groups[-1].append(BreakWire.fillup_layer(len(layer), '»'))

                # New group
                layer_groups.append([BreakWire.fillup_layer(len(layer), '«')])
                rest_of_the_line = line_length - layer_groups[-1][-1][0].length

                layer_groups[-1].append(
                    InputWire.fillup_layer(self.wire_names(with_initial_state=False)))
                rest_of_the_line -= layer_groups[-1][-1][0].length

                layer_groups[-1].append(layer)
                rest_of_the_line -= layer_groups[-1][-1][0].length

        lines = []

        for layer_group in layer_groups:
            wires = list(zip(*layer_group))
            lines += self.draw_wires(wires)

        return lines

    def wire_names(self, with_initial_state=False):
        """Returns a list of names for each wire.

        Args:
            with_initial_state (bool): Optional (Default: False). If true, adds
                the initial value to the name.

        Returns:
            List: The list of wire names.
        """
        if with_initial_state:
            initial_qubit_value = '|0>'
            # initial_clbit_value = '0 '
        else:
            initial_qubit_value = ''
            # initial_clbit_value = ''

        qubit_labels = []
        if self.layout is None:
            count = 0
            for bit in self.qregs:
                label = '{name}_{index}: ' + initial_qubit_value
                qubit_labels.append(label.format(name="q",
                                                 index=count,
                                                 physical=''))
                count += 1
        else:
            for bit in self.qregs:
                if self.layout[bit.index]:
                    label = '{name}_{index} -> {physical} ' + initial_qubit_value
                    qubit_labels.append(label.format(name=self.layout[bit.index].register.name,
                                                     index=self.layout[bit.index].index,
                                                     physical=bit.index))
                else:
                    qubit_labels.append('%s ' % bit.index + initial_qubit_value)

        return qubit_labels

    def should_compress(self, top_line, bot_line):
        """Decides if the top_line and bot_line should be merged,
        based on `self.vertical_compression`."""
        if self.vertical_compression == 'high':
            return True
        if self.vertical_compression == 'low':
            return False
        for top, bot in zip(top_line, bot_line):
            if top in ['┴', '╨'] and bot in ['┬', '╥']:
                return False
        for line in (bot_line, top_line):
            no_spaces = line.replace(' ', '')
            if len(no_spaces) > 0 and all(c.isalpha() or c.isnumeric() for c in no_spaces):
                return False
        return True

    def draw_wires(self, wires):
        """Given a list of wires, creates a list of lines with the text drawing.

        Args:
            wires (list): A list of wires with instructions.
        Returns:
            list: A list of lines with the text drawing.
        """
        lines = []
        bot_line = None
        for wire in wires:
            # TOP
            top_line = ''
            for instruction in wire:
                top_line += instruction.top

            if bot_line is None:
                lines.append(top_line)
            else:
                if self.should_compress(top_line, bot_line):
                    lines.append(TextDrawing.merge_lines(lines.pop(), top_line))
                else:
                    lines.append(TextDrawing.merge_lines(lines[-1], top_line, icod="bot"))

            # MID
            mid_line = ''
            for instruction in wire:
                mid_line += instruction.mid
            lines.append(TextDrawing.merge_lines(lines[-1], mid_line, icod="bot"))

            # BOT
            bot_line = ''
            for instruction in wire:
                bot_line += instruction.bot
            lines.append(TextDrawing.merge_lines(lines[-1], bot_line, icod="bot"))

        return lines

    @staticmethod
    def label_for_conditional(instruction):
        """ Creates the label for a conditional instruction."""
        return "= %s" % instruction.condition[1]

    @staticmethod
    def params_for_label(gate):
        """Get the params and format them to add them to a label. None if there
         are no params or if the params are numpy.ndarrays."""
        ret = []
        for param in gate.pargs:
            try:
                str_param = pi_check(param, ndigits=5)
                ret.append('%s' % str_param)
            except TypeError:
                ret.append('%s' % param)
        return ret

    @staticmethod
    def special_label(instruction):
        """Some instructions have special labels"""
        labels = {IGate: 'I',
                  Initialize: 'initialize',
                  UnitaryGate: 'unitary',
                  HamiltonianGate: 'Hamiltonian',
                  SXGate: '√X',
                  SXdgGate: '√XDG'}
        instruction_type = type(instruction)
        if instruction_type in {Gate, Instruction}:
            return instruction.qasm_name
        return labels.get(instruction_type, None)

    @staticmethod
    def label_for_box(instruction, controlled=False):
        """ Creates the label for a box."""
        label = instruction.qasm_name
        params = TextDrawing.params_for_label(instruction)
        if params:
            label += "(%s)" % ','.join(params)
        return label

    @staticmethod
    def merge_lines(top, bot, icod="top"):
        """Merges two lines (top and bot) in the way that the overlapping make senses.

        Args:
            top (str): the top line
            bot (str): the bottom line
            icod (top or bot): in case of doubt, which line should have priority? Default: "top".
        Returns:
            str: The merge of both lines.
        """
        ret = ""
        for topc, botc in zip(top, bot):
            if topc == botc:
                ret += topc
            elif topc in '┼╪' and botc == " ":
                ret += "│"
            elif topc == " ":
                ret += botc
            elif topc in '┬╥' and botc in " ║│" and icod == "top":
                ret += topc
            elif topc in '┬' and botc == " " and icod == "bot":
                ret += '│'
            elif topc in '╥' and botc == " " and icod == "bot":
                ret += '║'
            elif topc in '┬│' and botc == "═":
                ret += '╪'
            elif topc in '┬│' and botc == "─":
                ret += '┼'
            elif topc in '└┘║│░' and botc == " " and icod == "top":
                ret += topc
            elif topc in '─═' and botc == " " and icod == "top":
                ret += topc
            elif topc in '─═' and botc == " " and icod == "bot":
                ret += botc
            elif topc in "║╥" and botc in "═":
                ret += "╬"
            elif topc in "║╥" and botc in "─":
                ret += "╫"
            elif topc in '║╫╬' and botc in " ":
                ret += "║"
            elif topc == '└' and botc == "┌" and icod == 'top':
                ret += "├"
            elif topc == '┘' and botc == "┐":
                ret += "┤"
            elif botc in "┐┌" and icod == 'top':
                ret += "┬"
            elif topc in "┘└" and botc in "─" and icod == 'top':
                ret += "┴"
            elif botc == " " and icod == 'top':
                ret += topc
            else:
                ret += botc
        return ret

    @staticmethod
    def normalize_width(layer):
        """
        When the elements of the layer have different widths, sets the width to the max elements.

        Args:
            layer (list): A list of elements.
        """
        instructions = list(filter(lambda x: x is not None, layer))
        longest = max([instruction.length for instruction in instructions])
        for instruction in instructions:
            instruction.layer_width = longest

    @staticmethod
    def controlled_wires(instruction, layer):
        """
        Analyzes the instruction in the layer and checks if the controlled arguments are in
        the box or out of the box.

        Args:
            instruction (Instruction): instruction to analyse
            layer (Layer): The layer in which the instruction is inserted.

        Returns:
            Tuple(list, list, list):
              - tuple: controlled arguments on top of the "instruction box", and its status
              - tuple: controlled arguments on bottom of the "instruction box", and its status
              - tuple: controlled arguments in the "instruction box", and its status
              - the rest of the arguments
        """
        num_ctrl_qubits = instruction.targets
        ctrl_qubits = instruction.cargs
        args_qubits = instruction.targs
        ctrl_state = "{:b}".format(1).rjust(num_ctrl_qubits, '0')[::-1]

        in_box = list()
        top_box = list()
        bot_box = list()

        qubit_index = sorted([i for i in ctrl_qubits + args_qubits])

        for ctrl_qubit in zip(ctrl_qubits, ctrl_state):
            if min(qubit_index) > layer.qregs.index(ctrl_qubit[0]):
                top_box.append(ctrl_qubit)
            elif max(qubit_index) < layer.qregs.index(ctrl_qubit[0]):
                bot_box.append(ctrl_qubit)
            else:
                in_box.append(ctrl_qubit)
        return (top_box, bot_box, in_box, args_qubits)

    @staticmethod
    def resolution_layers(gates):
        layers = [circuit_layer()]
        for gate in gates:
            for i in range(len(layers) - 1, -2, -1):
                if i == -1 or not layers[i].checkGate(gate):
                    if i + 1 >= len(layers):
                        layers.append(circuit_layer())
                    layers[i + 1].addGate(gate)
                    break
        return layers

    def _set_ctrl_state(self, instruction, conditional, ctrl_label, bottom):
        """ Takes the ctrl_state from instruction and appends Bullet or OpenBullet
        to gates depending on whether the bit in ctrl_state is 1 or 0. Returns gates"""

        gates = []
        num_ctrl_qubits = instruction.controls
        ctrl_qubits = instruction.cargs
        cstate = "{:b}".format(1).rjust(num_ctrl_qubits, '0')[::-1]
        for i in range(len(ctrl_qubits)):
            if cstate[i] == '1':
                gates.append(Bullet(conditional=conditional, label=ctrl_label,
                                    bottom=bottom))
            else:
                gates.append(OpenBullet(conditional=conditional, label=ctrl_label,
                                        bottom=bottom))
        return gates

    def _instruction_to_gate(self, gate, layer):
        """ Convert an gates into its corresponding Gate object, and establish
        any connections it introduces between qubits"""

        current_cons = []
        connection_label = None
        ctrl_label = None
        box_label = None
        conditional = False
        # multi_qubit_instruction = gate.targets >= 2
        # label_multibox = False

        # add in a gate that operates over multiple qubits
        def add_connected_gate(gate, gates, layer, current_cons):
            for i, g in enumerate(gates):
                gate_args = gate.cargs + gate.targs
                actual_index = self.qregs.index(gate_args[i])
                if actual_index not in [i for i, j in current_cons]:
                    layer.set_qubit(gate_args[i], g)
                    current_cons.append((actual_index, g))

        ctrl_label = ""
        box_label = gate.qasm_name

        if isinstance(gate, MeasureGate):
            mgate = MeasureFrom()
            layer.set_qubit(gate.targs[0], mgate)
        elif isinstance(gate, BarrierGate):
            layer.set_qubit(gate.targ, Barrier())
        elif isinstance(gate, SwapGate):
            # swap
            gates = [Ex(conditional=conditional) for _ in range(len(gate.cargs + gate.targs))]
            add_connected_gate(gate, gates, layer, current_cons)

        elif isinstance(gate, ResetGate):
            # reset
            layer.set_qubit(gate.targs[0], Reset(conditional=conditional))

        elif isinstance(gate, RzzGate):
            # rzz
            connection_label = "ZZ(%s)" % TextDrawing.params_for_label(gate)[0]
            gates = [Bullet(conditional=conditional), Bullet(conditional=conditional)]
            add_connected_gate(gate, gates, layer, current_cons)

        elif gate.targets + gate.controls == 1:
            # unitary gate
            layer.set_qubit(gate.targ,
                            BoxOnQuWire(TextDrawing.label_for_box(gate),
                                        conditional=conditional))

        elif gate.controls >= 1:
            label = box_label if box_label is not None \
                else TextDrawing.label_for_box(gate, controlled=True)
            params_array = TextDrawing.controlled_wires(gate, layer)
            controlled_top, controlled_bot, controlled_edge, rest = params_array
            gates = self._set_ctrl_state(gate, conditional, ctrl_label,
                                         bool(controlled_bot))
            gates.append(BoxOnQuWire(label, conditional=conditional))
            add_connected_gate(gate, gates, layer, current_cons)

        elif gate.targets >= 2:
            # multiple qubit gate
            label = TextDrawing.label_for_box(gate)
            layer.set_qu_multibox(gate.targs, label, conditional=conditional)

        elif gate.targs:
            # multiple gate, involving both qargs AND cargs
            label = TextDrawing.label_for_box(gate)
            layer._set_multibox(label, qubits=gate.targs,
                                conditional=conditional)
        else:
            raise Exception(
                f"Text visualizer does not know how to handle this instruction: {gate.type}."
            )

        # sort into the order they were declared in
        # this ensures that connected boxes have lines in the right direction
        current_cons.sort(key=lambda tup: tup[0])
        current_cons = [g for q, g in current_cons]

        return layer, current_cons, connection_label

    def build_layers(self):
        """
        Constructs layers.
        Returns:
            list: List of DrawElements.
        Raises:
            VisualizationError: When the drawing is, for some reason, impossible to be drawn.
        """
        wire_names = self.wire_names(with_initial_state=self.initial_state)
        if not wire_names:
            return []

        layers = [InputWire.fillup_layer(wire_names)]

        layer = Layer(self.qregs)
        for gate_layer in self.layers:
            layer = Layer(self.qregs)
            for gate in gate_layer.gates:
                layer, current_connections, connection_label = \
                    self._instruction_to_gate(gate, layer)

                layer.connections.append((connection_label, current_connections))
            layer.connect_with("│")
            layers.append(layer.full_layer)
        return layers


class Layer:
    """ A layer is the "column" of the circuit. """

    def __init__(self, qregs):
        self.qregs = qregs
        self.qubit_layer = [None] * len(qregs)
        self.connections = []

    @property
    def full_layer(self):
        """
        Returns the composition of qubits and classic wires.
        Returns:
            String: self.qubit_layer + self.clbit_layer
        """
        return self.qubit_layer

    def set_qubit(self, qubit, element):
        """Sets the qubit to the element.

        Args:
            qubit (qbit): Element of self.qregs.
            element (DrawElement): Element to set in the qubit
        """
        self.qubit_layer[self.qregs.index(qubit)] = element

    def _set_multibox(self, label, qubits=None, top_connect=None,
                      bot_connect=None, conditional=False, controlled_edge=None):
        bits = list(qubits)
        bit_index = sorted([i for i in self.qregs if i in bits])
        wire_label_len = len(str(len(bits) - 1))
        qargs = [str(bits.index(qbit)).ljust(wire_label_len, ' ')
                 for qbit in self.qregs if qbit in bits]
        bits.sort()
        set_bit = self.set_qubit
        OnWire = BoxOnQuWire
        OnWireTop = BoxOnQuWireTop
        OnWireMid = BoxOnQuWireMid
        OnWireBot = BoxOnQuWireBot

        control_index = {}
        if controlled_edge:
            for index, qubit in enumerate(self.qregs):
                for qubit_in_edge, value in controlled_edge:
                    if qubit == qubit_in_edge:
                        control_index[index] = '■' if value == '1' else 'o'
        if len(bit_index) == 1:
            set_bit(bits[0], OnWire(label, top_connect=top_connect))
        else:
            box_height = max(bit_index) - min(bit_index) + 1
            set_bit(bits.pop(0), OnWireTop(label, top_connect=top_connect, wire_label=qargs.pop(0)))
            for order, bit_i in enumerate(range(min(bit_index) + 1, max(bit_index))):
                if bit_i in bit_index:
                    named_bit = bits.pop(0)
                    wire_label = qargs.pop(0)
                else:
                    named_bit = (self.qregs)[bit_i]
                    wire_label = ' ' * wire_label_len

                control_label = control_index.get(bit_i)
                set_bit(named_bit, OnWireMid(label, box_height, order, wire_label=wire_label,
                                             control_label=control_label))
            set_bit(bits.pop(0), OnWireBot(label, box_height, bot_connect=bot_connect,
                                           wire_label=qargs.pop(0), conditional=conditional))
        return bit_index

    def set_qu_multibox(self, bits, label, top_connect=None, bot_connect=None,
                        conditional=False, controlled_edge=None):
        """Sets the multi qubit box.

        Args:
            bits (list[int]): A list of affected bits.
            label (string): The label for the multi qubit box.
            top_connect (char): None or a char connector on the top
            bot_connect (char): None or a char connector on the bottom
            conditional (bool): If the box has a conditional
            controlled_edge (list): A list of bit that are controlled (to draw them at the edge)
        Return:
            List: A list of indexes of the box.
        """
        return self._set_multibox(label, qubits=bits, top_connect=top_connect,
                                  bot_connect=bot_connect,
                                  conditional=conditional, controlled_edge=controlled_edge)

    def connect_with(self, wire_char):
        """Connects the elements in the layer using wire_char.

        Args:
            wire_char (char): For example '║' or '│'.
        """

        if len([qbit for qbit in self.qubit_layer if qbit is not None]) == 1:
            # Nothing to connect
            return

        for label, affected_bits in self.connections:

            if not affected_bits:
                continue

            affected_bits[0].connect(wire_char, ['bot'])
            for affected_bit in affected_bits[1:-1]:
                affected_bit.connect(wire_char, ['bot', 'top'])

            affected_bits[-1].connect(wire_char, ['top'], label)

            if label:
                for affected_bit in affected_bits:
                    affected_bit.right_fill = len(label) + len(affected_bit.mid)



"""
Compute the disentangler of Pauli operators (O, O')
"""

import copy
import random
import itertools

import numpy as np

from QuICT.core.gate import build_gate, BasicGate, CompositeGate, GateType, PAULI_GATE_SET, CX, H


class PauliOperator(object):
    """
    Pauli operator is a list of (I, X, Y, Z) with length n, which operates on n qubits.

    In this class, we use a list of GateType to represent the operator, where the GateTypes stand
    for the gates. Despite the operator, the phase is also recorded.
    """
    pauli_list = [GateType.id, GateType.x, GateType.y, GateType.z]
    phase_list = [1 + 0j, 0 + 1j, -1 + 0j, 0 - 1j]

    combine_rules = {
        # I I = I
        (GateType.id, GateType.id): (GateType.id, 1),
        # I X = X
        (GateType.id, GateType.x): (GateType.x, 1),
        # I Y = Y
        (GateType.id, GateType.y): (GateType.y, 1),
        # I Z = Z
        (GateType.id, GateType.z): (GateType.z, 1),
        # X I = X
        (GateType.x, GateType.id): (GateType.x, 1),
        # X X = I
        (GateType.x, GateType.x): (GateType.id, 1),
        # X Y = -iZ
        (GateType.x, GateType.y): (GateType.z, -1j),
        # X Z = iY
        (GateType.x, GateType.z): (GateType.y, 1j),
        # Y I = Y
        (GateType.y, GateType.id): (GateType.y, 1),
        # Y X = iZ
        (GateType.y, GateType.x): (GateType.z, 1j),
        # Y Y = I
        (GateType.y, GateType.y): (GateType.id, 1),
        # Y Z = -iX
        (GateType.y, GateType.z): (GateType.x, -1j),
        # Z I = Z
        (GateType.z, GateType.id): (GateType.z, 1),
        # Z X = -iY
        (GateType.z, GateType.x): (GateType.y, -1j),
        # Z Y = iX
        (GateType.z, GateType.y): (GateType.x, 1j),
        # Z Z = I
        (GateType.z, GateType.z): (GateType.id, 1)
    }

    conjugate_rules = {
        # CX I0 I1 CX = I0 I1
        (GateType.cx, GateType.id, GateType.id): (GateType.id, GateType.id, 1),
        # CX I0 X1 CX = I0 X1
        (GateType.cx, GateType.id, GateType.x): (GateType.id, GateType.x, 1),
        # CX I0 Y1 CX = Z0 Y1
        (GateType.cx, GateType.id, GateType.y): (GateType.z, GateType.y, 1),
        # CX I0 Z1 CX = Z0 Z1
        (GateType.cx, GateType.id, GateType.z): (GateType.z, GateType.z, 1),
        # CX X0 I1 CX = X0 X1
        (GateType.cx, GateType.x, GateType.id): (GateType.x, GateType.x, 1),
        # CX X0 X1 CX = X0 I1
        (GateType.cx, GateType.x, GateType.x): (GateType.x, GateType.id, 1),
        # CX X0 Y1 CX = Y0 Z1
        (GateType.cx, GateType.x, GateType.y): (GateType.y, GateType.z, 1),
        # CX X0 Z1 CX = -Y0 Y1
        (GateType.cx, GateType.x, GateType.z): (GateType.y, GateType.y, -1),
        # CX Y0 I1 CX = Y0 X1
        (GateType.cx, GateType.y, GateType.id): (GateType.y, GateType.x, 1),
        # CX Y0 X1 CX = Y0 I1
        (GateType.cx, GateType.y, GateType.x): (GateType.y, GateType.id, 1),
        # CX Y0 Y1 CX = -X0 Z1
        (GateType.cx, GateType.y, GateType.y): (GateType.x, GateType.z, -1),
        # CX Y0 Z1 CX = X0 Y1
        (GateType.cx, GateType.y, GateType.z): (GateType.x, GateType.y, 1),
        # CX Z0 I1 CX = Z0 I1
        (GateType.cx, GateType.z, GateType.id): (GateType.z, GateType.id, 1),
        # CX Z0 X1 CX = Z0 X1
        (GateType.cx, GateType.z, GateType.x): (GateType.z, GateType.x, 1),
        # CX Z0 Y1 CX = I0 Y1
        (GateType.cx, GateType.z, GateType.y): (GateType.id, GateType.y, 1),
        # CX Z0 Z1 CX = I0 Z1
        (GateType.cx, GateType.z, GateType.z): (GateType.id, GateType.z, 1),
        # H I H = I
        (GateType.h, GateType.id): (GateType.id, 1),
        # H X H = Z
        (GateType.h, GateType.x): (GateType.z, 1),
        # H Y H = -Y
        (GateType.h, GateType.y): (GateType.y, -1),
        # H Z H = X
        (GateType.h, GateType.z): (GateType.x, 1),
        # Sdg I S = I
        (GateType.s, GateType.id): (GateType.id, 1),
        # Sdg X S = Y
        (GateType.s, GateType.x): (GateType.y, 1),
        # Sdg Y S = -X
        (GateType.s, GateType.y): (GateType.x, -1),
        # Sdg Z S = Z
        (GateType.s, GateType.z): (GateType.z, 1),
        # S I Sdg = I
        (GateType.sdg, GateType.id): (GateType.id, 1),
        # S X Sdg = -Y
        (GateType.sdg, GateType.x): (GateType.y, -1),
        # S Y Sdg = X
        (GateType.sdg, GateType.y): (GateType.x, 1),
        # S Z Sdg = Z
        (GateType.sdg, GateType.z): (GateType.z, 1),
        # X I X = I
        (GateType.x, GateType.id): (GateType.id, 1),
        # X X X = X
        (GateType.x, GateType.x): (GateType.x, 1),
        # X Y X = -Y
        (GateType.x, GateType.y): (GateType.y, -1),
        # X Z X = -Z
        (GateType.x, GateType.z): (GateType.z, -1),
        # Y I Y = I
        (GateType.y, GateType.id): (GateType.id, 1),
        # Y X Y = -X
        (GateType.y, GateType.x): (GateType.x, -1),
        # Y Y Y = Y
        (GateType.y, GateType.y): (GateType.y, 1),
        # Y Z Y = -Z
        (GateType.y, GateType.z): (GateType.z, -1),
        # Z I Z = I
        (GateType.z, GateType.id): (GateType.id, 1),
        # Z X Z = -X
        (GateType.z, GateType.x): (GateType.x, -1),
        # Z Y Z = -Y
        (GateType.z, GateType.y): (GateType.y, -1),
        # Z Z Z = Z
        (GateType.z, GateType.z): (GateType.z, 1)
    }

    standardize_rules = {
        (GateType.id, GateType.id): ([], 4),
        (GateType.id, GateType.x): ([GateType.h], 3),
        (GateType.id, GateType.y): ([GateType.s, GateType.h], 3),
        (GateType.id, GateType.z): ([], 3),
        (GateType.x, GateType.id): ([], 2),
        (GateType.x, GateType.x): ([], 1),
        (GateType.x, GateType.y): ([GateType.h, GateType.s, GateType.h], 0),
        (GateType.x, GateType.z): ([], 0),
        (GateType.y, GateType.id): ([GateType.s], 2),
        (GateType.y, GateType.x): ([GateType.h, GateType.s], 0),
        (GateType.y, GateType.y): ([GateType.s], 1),
        (GateType.y, GateType.z): ([GateType.s], 0),
        (GateType.z, GateType.id): ([GateType.h], 2),
        (GateType.z, GateType.x): ([GateType.h], 0),
        (GateType.z, GateType.y): ([GateType.s, GateType.h], 0),
        (GateType.z, GateType.z): ([GateType.h], 1)
    }

    def __init__(self, operator=None, phase=1 + 0j):
        """
        Construct a PauliOperator with a list of GateType

        Args:
            operator(list, optional): the list of GateType representing the PauliOperator
            phase(complex, optional): the global phase of the PauliOperator, ±1 or ±i
        """
        if operator is None:
            self._operator = []
        else:
            assert isinstance(operator, list), TypeError("operator must be list of int(GateType).")
            for gate_type in operator:
                assert gate_type == GateType.id or gate_type in PAULI_GATE_SET,\
                    ValueError("operator must contain Pauli gates only.")
            self._operator = operator
        assert phase in PauliOperator.phase_list, ValueError("phase must be ±1 or ±i")
        self._phase = phase

    def __str__(self):
        string = ''
        for i in range(len(self.operator)):
            if self.operator[i] == GateType.id:
                string += 'I'
            elif self.operator[i] == GateType.x:
                string += 'X'
            elif self.operator[i] == GateType.y:
                string += 'Y'
            elif self.operator[i] == GateType.z:
                string += 'Z'
            else:
                raise ValueError('Invalid GateType')
        return string + ', {}'.format(self.phase)

    @property
    def operator(self) -> list:
        return self._operator

    @operator.setter
    def operator(self, operator: list):
        assert isinstance(operator, list), TypeError("operator must be list of int(GateType).")
        for gate_type in operator:
            assert gate_type == GateType.id or gate_type in PAULI_GATE_SET,\
                ValueError("operator must contain Pauli gates only.")
        self._operator = operator

    @property
    def phase(self) -> complex:
        return self._phase

    @phase.setter
    def phase(self, phase: complex):
        assert phase in PauliOperator.phase_list, ValueError("phase must be ±1 or ±i")
        self._phase = phase

    def gates(self, keep_id=False, keep_phase=False) -> CompositeGate:
        """
        The CompositeGate corresponding to the PauliOperator

        Args:
            keep_id(bool, optional): whether to keep the IDgate in the CompositeGate
            keep_phase(bool, optional): whether to keep the global phase in the CompositeGate

        Returns:
            CompositeGate: The CompositeGate corresponding to the PauliOperator
        """
        gates = CompositeGate()
        for qubit, gate_type in enumerate(self.operator):
            if not keep_id and gate_type == GateType.id:
                continue
            gate = build_gate(gate_type, qubit)
            gates.append(gate)
        if keep_phase and not np.isclose(self.phase, 1 + 0j):
            phase = -1j * np.log(self.phase)
            gate = build_gate(GateType.gphase, 0, phase.real)
            gates.append(gate)
        return gates

    @property
    def width(self) -> int:
        return len(self.operator)

    @property
    def support(self) -> list:
        support = []
        for qubit, operator in enumerate(self.operator):
            if operator != GateType.id:
                support.append(qubit)
        return support

    @property
    def hamming_weight(self) -> int:
        return len(self.support)

    @staticmethod
    def random(width: int):
        """
        Give a random PauliOperator with given width

        Args:
            width(int): the width of the PauliOperator

        Returns:
            PauliOperator: a random PauliOperator with given width
        """
        operator = []
        for _ in range(width):
            operator.append(random.choice(PauliOperator.pauli_list))
        return PauliOperator(operator)

    @staticmethod
    def random_anti_commutative_pair(width):
        op_id = [GateType.id for _ in range(width)]
        # Avoid p1 = identity
        while True:
            p1 = PauliOperator.random(width)
            if p1.operator != op_id:
                break
        # Expectation = 2
        while True:
            p2 = PauliOperator.random(width)
            if not p1.commute(p2):
                break
        return p1, p2

    @staticmethod
    def iterator(width: int):
        """
        Yield all the PauliOperators with given width

        Args:
            width(int): the width of the PauliOperator

        Yields:
            PauliOperator: PauliOperator with given width
        """
        for operator in itertools.product(PauliOperator.pauli_list, repeat=width):
            yield PauliOperator(list(operator))

    def commute(self, other):
        """
        Decide whether two PauliOperators commute or anti-commute

        Args:
            other(PauliOperator): PauliOperator to be checked with self

        Returns:
            boolean: True for commutative or False for anti-commutative
        """
        assert isinstance(other, PauliOperator),\
            TypeError("commute() only checks commutativity between PauliOperators.")
        assert self.width == other.width,\
            ValueError("PauliOperators to be checked must have the same width.")

        res = True
        for qubit in range(self.width):
            if self.operator[qubit] == GateType.id or other.operator[qubit] == GateType.id\
               or self.operator[qubit] == other.operator[qubit]:
                continue
            else:
                res = not res
        return res

    def combine(self, other):
        """
        Compute the PauliOperator after combined with a PauliOperator from the right side.
        Be aware of the order, which would affect the global phase.

        Args:
            other(PauliOperator): the PauliOperator to be combined

        Returns:
            PauliOperator: the combined PauliOperator
        """
        assert isinstance(other, PauliOperator),\
            TypeError("combine() only combines PauliOperators.")
        assert self.width == other.width,\
            ValueError("PauliOperators to be combined must have the same width.")

        self.phase *= other.phase
        for qubit, gate_type in enumerate(other.operator):
            self.combine_one_gate(gate_type, qubit)
        return self

    def combine_one_gate(self, gate_type: GateType, qubit: int):
        """
        Compute the PauliOperator after combined with a Pauli gate from the right side.
        Be aware of the order, which would affect the global phase.

        Args:
            gate_type(GateType): type of the Pauli gate to be combined
            qubit(int): qubit that the Pauli gate acts on
        """
        assert gate_type in PAULI_GATE_SET or gate_type == GateType.id,\
            ValueError('gate to be combined must be a pauli gate')
        assert isinstance(qubit, int) and qubit >= 0 and qubit < self.width,\
            ValueError('qubit out of range')

        self.operator[qubit], phase = PauliOperator.combine_rules[self.operator[qubit], gate_type]
        self.phase *= phase

    def conjugate_act(self, gate: BasicGate):
        """
        Compute the PauliOperator after conjugate action of a clifford gate
        Be aware that the conjugate action means U^-1 P U, where U is the clifford gate
        and P is the PauliOperator. It is important for S gate.

        Args:
            gate(BasicGate): the clifford gate to be acted on the PauliOperator
        """
        assert gate.is_clifford() or gate.type == GateType.id,\
            ValueError("Only conjugate action of Clifford gates here.")
        for targ in gate.cargs + gate.targs:
            assert targ < self.width, ValueError("target of the gate out of range")

        if gate.type == GateType.cx:
            self.operator[gate.carg], self.operator[gate.targ], phase\
                = PauliOperator.conjugate_rules[gate.type, self.operator[gate.carg], self.operator[gate.targ]]
        else:
            self.operator[gate.targ], phase = PauliOperator.conjugate_rules[gate.type, self.operator[gate.targ]]
        self.phase *= phase

    @staticmethod
    def disentangler(pauli_x, pauli_z, target=0) -> CompositeGate:
        """
        For anti-commuting n-qubit Pauli operators O and O', there exists a Clifford circuit L∈C_n
        such that L^(-1) O L = X_j, L^(-1) O' L = Z_j.
        L is referred to as a disentangler for the pair (O, O').

        Args:
            pauli_x(PauliOperator): the PauliOperator to be transformed to X_j
            pauli_z(PauliOperator): the PauliOperator to be transformed to Z_j
            target(int, optional): the j in the X_j, Z_j to be transformed to

        Returns:
            CompositeGate: the disentangler for the pair (O, O')

        Reference:
            https://arxiv.org/abs/2105.02291
        """
        assert isinstance(pauli_x, PauliOperator) and isinstance(pauli_z, PauliOperator),\
            TypeError("disentangler only defined for PauliOperator pairs")
        assert pauli_x.width == pauli_z.width,\
            ValueError('two PauliOperators must be of the same width')
        assert not pauli_x.commute(pauli_z),\
            ValueError('anti-commutative pairs needed for computing disentangler')
        assert isinstance(target, int) and 0 <= target and target < pauli_x.width,\
            ValueError('the target must be integer in the width of the PauliOperators')

        pauli_x = copy.deepcopy(pauli_x)
        pauli_z = copy.deepcopy(pauli_z)

        # Record the qubits in the 5 cases of the standard forms
        XZ = []
        XX = []
        XI = []
        IZ = []
        II = []
        standard_forms = [XZ, XX, XI, IZ, II]

        # Transform the pauli_x and pauli_z to 'standard' forms
        standardizer = CompositeGate()
        for qubit in range(pauli_x.width):
            gates, form = PauliOperator.standardize_rules[pauli_x.operator[qubit], pauli_z.operator[qubit]]
            for gate_type in gates:
                gate = build_gate(gate_type, qubit)
                pauli_x.conjugate_act(gate)
                pauli_z.conjugate_act(gate)
                standardizer.append(gate)
            standard_forms[form].append(qubit)

        # Construct the disentangler of 'standard' pairs with the algorithm given in reference
        disentangler = CompositeGate()
        with disentangler:
            for subset in standard_forms:
                if target in subset:
                    if subset is not XZ:
                        # Swap & [target, XZ[0]]
                        CX & [target, XZ[0]]
                        CX & [XZ[0], target]
                        CX & [target, XZ[0]]
                    if XZ[0] != target:
                        subset.remove(target)
                        subset.append(XZ[0])
                        XZ[0] = target
            for j in XI:
                CX & [target, j]
            for j in IZ:
                CX & [j, target]
            if len(XX) > 0:
                i = XX[0]
                for j in XX[1:]:
                    CX & [i, j]
                CX & [target, i]
                H & i
                CX & [i, target]
            # Anti-commutation gives that len(XZ) is odd
            for k in range((len(XZ) - 1) // 2):
                i = 2 * k + 1
                j = 2 * k + 2
                CX & [XZ[j], XZ[i]]
                CX & [XZ[i], target]
                CX & [target, XZ[j]]

        standardizer.extend(disentangler)

        # Add the last Pauli gate for the phase correction
        for gate in disentangler:
            pauli_x.conjugate_act(gate)
            pauli_z.conjugate_act(gate)
        if pauli_x.phase == 1 and pauli_z.phase == -1:
            gate = build_gate(GateType.x, target)
            standardizer.append(gate)
        if pauli_x.phase == -1 and pauli_z.phase == -1:
            gate = build_gate(GateType.y, target)
            standardizer.append(gate)
        if pauli_x.phase == -1 and pauli_z.phase == 1:
            gate = build_gate(GateType.z, target)
            standardizer.append(gate)

        return standardizer



import os
from os import path as osp
from typing import Dict, List, Tuple, Union

import networkx as nx
from numba import njit
import numpy as np

from QuICT.core import *


@njit
def _floyd(n: int, dist: np.ndarray, _inf: int) -> np.ndarray:
    for i in range(n):
        dist[i][i] = 0
    for k in range(n):
        for i in range(n):
            for j in range(n):
                dist[i][j] = min(dist[i][j], dist[i][k] + dist[k][j])
    for i in range(n):
        for j in range(n):
            if dist[i][j] >= _inf:
                dist[i][j] = 0
    return dist


class DataFactory:
    def __init__(
        self,
        topo: Union[str, Layout],
        max_gate_num: int,
        data_dir: str = None,
    ) -> None:
        if data_dir is None:
            data_dir = osp.dirname(osp.realpath(__file__))
            data_dir = osp.join(data_dir, "data")
        self._data_dir = data_dir
        self._topo_dir = osp.join(data_dir, "topo")

        self._max_gate_num = max_gate_num

        self._topo_names = None

        # Topo attr cache def
        # These attributes maps are lazily initialized for faster start up.
        self._topo_map = {}
        self._topo_graph_map = {}
        self._topo_edge_map = {}
        self._topo_edge_mat_map = {}
        self._topo_qubit_num_map = {}
        self._topo_dist_map = {}

        if isinstance(topo, str):
            self._cur_topo = self.topo_map[topo]
        elif isinstance(topo, Layout):
            self._cur_topo = topo
        else:
            raise TypeError("Only support layout name or Layout object.")
        assert self._cur_topo is not None

    @property
    def topo_names(self) -> List[str]:
        if self._topo_names is None:
            self._topo_names = []
            for _, _, filenames in os.walk(self._topo_dir):
                for name in filenames:
                    self._topo_names.append(name.split(".")[0])
        return self._topo_names

    @property
    def topo_graph_map(self) -> Dict[str, nx.Graph]:
        if len(self._topo_graph_map) == 0:
            self._reset_attr_cache()
        return self._topo_graph_map

    @property
    def topo_edge_map(self) -> Dict[str, List[Tuple[int, int]]]:
        if len(self._topo_edge_map) == 0:
            self._reset_attr_cache()
        return self._topo_edge_map

    @property
    def topo_qubit_num_map(self) -> Dict[str, int]:
        if len(self._topo_qubit_num_map) == 0:
            self._reset_attr_cache()
        return self._topo_qubit_num_map

    @property
    def topo_dist_map(self) -> Dict[str, np.ndarray]:
        if len(self._topo_dist_map) == 0:
            self._reset_attr_cache()
        return self._topo_dist_map

    @property
    def topo_edge_mat_map(self) -> Dict[str, np.ndarray]:
        if len(self._topo_edge_mat_map) == 0:
            self._reset_attr_cache()
        return self._topo_edge_mat_map

    @property
    def topo_map(self) -> Dict[str, Layout]:
        if len(self._topo_map) == 0:
            self._reset_attr_cache()
        return self._topo_map

    def _reset_attr_cache(self):
        for topo_name in self.topo_names:
            topo_path = osp.join(self._topo_dir, f"{topo_name}.json")
            topo = Layout.load_file(topo_path)
            self._topo_map[topo_name] = topo
            topo_graph = self.get_topo_graph(topo)
            self._topo_graph_map[topo_name] = topo_graph
            self._topo_qubit_num_map[topo_name] = topo.qubit_number
            self._topo_dist_map[topo_name] = self.get_topo_dist(topo_graph=topo_graph)

            topo_edge = self.get_topo_edges(topo_graph=topo_graph)
            topo_adj_mat_thin = np.zeros(
                (topo.qubit_number, topo.qubit_number), dtype=int
            )
            for u, v in topo_graph.edges:
                topo_adj_mat_thin[u][v] = 1
            self._topo_edge_map[topo_name] = topo_edge
            self._topo_edge_mat_map[topo_name] = topo_adj_mat_thin

    @classmethod
    def get_topo_graph(cls, topo: Layout) -> nx.Graph:
        """Build tha graph representation of a topology.

        Args:
            topo (Layout): Topology to be built.

        Returns:
            nx.Graph: Graph representation.
        """
        g = nx.Graph()
        for i in range(topo.qubit_number):
            g.add_node(i)
        for edge in topo:
            g.add_edge(edge.u, edge.v)
        return g

    @classmethod
    def get_topo_dist(cls, topo_graph: nx.Graph) -> np.ndarray:
        _inf = nx.number_of_nodes(topo_graph) + 5
        n = nx.number_of_nodes(topo_graph)
        dist = np.empty((n, n), dtype=np.int)
        dist[:, :] = _inf
        for u, v in topo_graph.edges:
            dist[u][v] = 1
            dist[v][u] = 1
        dist = _floyd(n, dist, _inf)
        return dist

    @classmethod
    def get_topo_edges(cls, topo_graph: nx.Graph) -> List[Tuple[int, int]]:
        topo_edge = []
        for u, v in topo_graph.edges:
            topo_edge.append((u, v))
        return topo_edge



from .. import InstructionSet
from ..transform_rule import u3_rule

from QuICT.core.gate import *

OriginSet = InstructionSet(
    GateType.cx,
    [GateType.u3]
)
OriginSet.register_one_qubit_rule(u3_rule)



import os
import sys

from QuICT.core import Layout
from QuICT.tools import Logger, LogFormat
from QuICT.tools.interface import OPENQASMInterface
from QuICT.tools.circuit_library import CircuitLib
from QuICT.qcda.qcda import QCDA
from QuICT.qcda.synthesis.gate_transform import USTCSet, GoogleSet, IBMQSet, IonQSet, NamSet, OriginSet
from QuICT.qcda.synthesis import GateTransform, CliffordUnidirectionalSynthesizer
from QuICT.qcda.optimization import (
    CliffordRzOptimization, CommutativeOptimization, SymbolicCliffordOptimization,
    TemplateOptimization, CnotWithoutAncilla
)


logger = Logger("QCDA_Local_Mode", LogFormat.full)


iset_mapping = {
    "USTC": USTCSet,
    "Google": GoogleSet,
    "IBMQ": IBMQSet,
    "IonQ": IonQSet,
    "Nam": NamSet,
    "Origin": OriginSet
}


def qcda_start(
    circuit_path: str,
    output_path: str,
    layout_path: str = None,
    methods: str = None,
    instruction_set: str = "Google",
    auto_mode: str = "light",
    para: bool = True,
    depara: bool = False,
    templates: str = ""
):
    if templates:
        width, size, depth = templates.split("+")
        templates = CircuitLib().get_template_circuit(
            int(width), int(size), int(depth)
        )

    method_mapping = {
        "GateTransform": GateTransform(iset_mapping[instruction_set]),
        "Clifford": CliffordUnidirectionalSynthesizer(),
        "CliffordRz": CliffordRzOptimization(auto_mode),
        "Commutative": CommutativeOptimization(para, depara),
        "SymbolicClifford": SymbolicCliffordOptimization(),
        "Template": TemplateOptimization(templates),
        "CNOT": CnotWithoutAncilla()
    }
    logger.debug("Start Run QCDA Job in local mode.")
    logger.debug(
        f"Job Parameters: circuit path: {circuit_path}, methods: {methods}, " +
        f"output path: {output_path}."
    )
    # Get circuit from given path
    circuit = OPENQASMInterface.load_file(circuit_path).circuit
    qcda = QCDA()

    if methods is not None:
        methods = methods.split("+")
        for method in methods:
            qcda.add_method(method_mapping[method])

    if layout_path is not None:
        layout = Layout.load_file(layout_path)
        qcda.add_mapping(layout)

    circuit_opt = qcda.compile(circuit)
    output_path = os.path.join(output_path, 'opt_circuit.qasm')
    circuit_opt.qasm(output_path)

    logger.debug(f"QCDA Job finished, store the result in {output_path}.")


if __name__ == "__main__":
    raw_args = sys.argv[1:]
    dict_args = dict([arg.split('=', maxsplit=1) for arg in raw_args])

    qcda_start(
        **dict_args
    )



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/12 11:35
# @Author  : Han Yu
# @File    : __init__.py.py

from .mct_one_aux import MCTOneAux
from .mct_linear_dirty_aux import MCTLinearHalfDirtyAux, MCTLinearOneDirtyAux
from .mct_without_aux import MCTWithoutAux



from typing import Union

from QuICT.core import *
from QuICT.core.gate import CompositeGate
from QuICT.core.gate.gate import *
from QuICT.qcda.utility import OutputAligner

from ..common import CircuitInfo, LayoutInfo
from .mcts_tree import MCTSTree


class MCTSMapping:
    def __init__(
        self,
        layout: Layout,
        bp_num: int = 10,
        sim_cnt: int = 10,
        sim_gate_num: int = 30,
        gamma: float = 0.9,
        epsilon: float = 0.001,
        c: float = 0.01,
    ) -> None:
        self._layout_info = LayoutInfo(layout=layout)
        self._bp_num = bp_num
        self._sim_cnt = sim_cnt
        self._sim_gate_num = sim_gate_num
        self._c = c
        self._gamma = gamma
        self._epsilon = epsilon
        self.logic2phy = []
        self.phy2logic = []

    @OutputAligner()
    def execute(
        self,
        circuit_like: Union[Circuit, CompositeGate],
    ) -> Union[Circuit, CompositeGate]:
        cg = CompositeGate()
        q = circuit_like.width()
        circuit_info = CircuitInfo(
            circ=circuit_like, max_gate_num=len(circuit_like.gates)
        )
        # Bypass first part
        circuit_info.eager_exec(
            logic2phy=list(range(q)),
            physical_circ=cg,
            topo_graph=self._layout_info.topo_graph,
        )
        # Execute search on remain
        mcts_tree = MCTSTree(
            circuit_info=circuit_info,
            layout_info=self._layout_info,
            bp_num=self._bp_num,
            sim_cnt=self._sim_cnt,
            sim_gate_num=self._sim_gate_num,
            gamma=self._gamma,
            epsilon=self._epsilon,
            c=self._c,
        )

        terminated = mcts_tree._root.is_terminated_node()
        while not terminated:
            action, part, terminated = mcts_tree.step()
            with cg:
                Swap & list(action)
            cg.extend(part)
        self.logic2phy, self.phy2logic = (
            mcts_tree._root.logic2phy,
            mcts_tree._root.phy2logic,
        )
        del mcts_tree

        result = cg
        return result



from flask import Blueprint

from .client.node_job_manage import QuICTLocalManager


job_blueprint = Blueprint(name="jobs", import_name=__name__)
URL_PREFIX = "/quict/jobs"

# Local Controller
local_job_manager = QuICTLocalManager()


@job_blueprint.route(f"{URL_PREFIX}/start", methods=["POST"])
def start_job(**kwargs):
    """start a job. """
    local_job_manager.start_job(kwargs["file"])

    return True


@job_blueprint.route(f"{URL_PREFIX}/<job_name>:delete", methods=["DELETE"])
def delete_job(job_name: str):
    """ delete a job. """
    local_job_manager.delete_job(job_name)


@job_blueprint.route(f"{URL_PREFIX}/<job_name>:status", methods=["GET"])
def status_jobs(job_name: str):
    return local_job_manager.delete_job(job_name)



"""
Class for customizing the whole process of synthesis, optimization and mapping
"""


from QuICT.qcda.synthesis import GateTransform
from QuICT.qcda.optimization import CommutativeOptimization, CliffordRzOptimization
from QuICT.qcda.mapping import MCTSMapping, SABREMapping
from QuICT.tools import Logger


logger = Logger("QCDA")


class QCDA(object):
    """ Customize the process of synthesis, optimization and mapping

    In this class, we meant to provide the users with a direct path to design the
    process by which they could transform a unitary matrix to a quantum circuit
    and/or optimize a quantum circuit.
    """

    def __init__(self, process=None):
        """ Initialize a QCDA process

        A QCDA process is defined by a list of synthesis, optimization and mapping.
        Experienced users could customize the process for certain purposes.

        Args:
            instruction(InstructionSet, optional): InstructionSet for default process
            layout(Layout, optional): Layout for default process
            process(list, optional): A customized list of Synthesis, Optimization and Mapping
        """
        self.process = []
        if process is not None:
            self.process = process

    def add_method(self, method=None):
        """ Adding a specific method to the process

        Args:
            method: Some QCDA method
        """
        self.process.append(method)

    def add_gate_transform(self, target_instruction=None):
        """ Add GateTransform for some target InstructionSet

        GateTransform would transform the gates in the original Circuit/CompositeGate to a certain InstructionSet.

        Args:
            instruction(InstructionSet): The target InstructionSet
        """
        assert target_instruction is not None, ValueError('No InstructionSet provided for Synthesis')
        self.add_method(GateTransform(target_instruction))

    def add_default_optimization(self, level='light'):
        """ Generate the default optimization process

        The default optimization process contains the CommutativeOptimization.

        Args:
            level(str): Optimizing level. Support `light`, `heavy` level.
        """

        self.add_method(CommutativeOptimization())
        self.add_method(CliffordRzOptimization(level=level))

    def add_mapping(self, layout=None, method='sabre'):
        """ Generate the default mapping process

        The default mapping process contains the Mapping

        Args:
            layout(Layout): Topology of the target physical device
            method(str, optional): used mapping method in ['mcts', 'sabre']
        """
        assert layout is not None, ValueError('No Layout provided for Mapping')
        assert method in ['mcts', 'sabre'], ValueError('Invalid mapping method')
        mapping_dict = {
            'mcts': MCTSMapping(layout=layout),
            'sabre': SABREMapping(layout=layout)
        }
        self.add_method(mapping_dict[method])

    def compile(self, circuit):
        """ Compile the circuit with the given process

        Args:
            circuit(CompositeGate/Circuit): the target CompositeGate or Circuit

        Returns:
            CompositeGate/Circuit: the resulting CompositeGate or Circuit
        """
        logger.info("QCDA Now processing GateDecomposition.")
        circuit.gate_decomposition()
        for process in self.process:
            logger.info(f"QCDA Now processing {process.__class__.__name__}.")
            circuit = process.execute(circuit)

        return circuit



from typing import Union
import numpy as np


CGATE_LIST = []


def matrix_product_to_circuit(
    gate_matrix: np.ndarray,
    gate_args: Union[int, list],
    max_q: int,
    min_q: int = 0,
    gpu_output: bool = False
):
    """ Expand gate matrix with the number of qubits

    Args:
        gate_matrix (np.ndarray): The gate's matrix.
        gate_args Union[int, list]: The gate's qubit indexes.
        max_q (int): The qubits' number
        min_q (int, optional): The minimum qubit's number. Defaults to 0.
        gpu_output(bool, optional): Generate matrix in GPU or not. Default to False.

    Returns:
        np.array: the expanded gate's 2-D matrix
    """
    if isinstance(gate_args, int):
        gate_args = [gate_args]

    n = 1 << (max_q - min_q)
    xor = n - 1
    new_values = np.zeros((n, n), dtype=gate_matrix.dtype)
    assert gate_matrix.shape == (1 << len(gate_args), 1 << len(gate_args))
    for arg in gate_args:
        assert arg >= 0 and arg < max_q and isinstance(arg, int)

    datas = np.zeros(n, dtype=int)
    for i in range(n):
        nowi = 0
        for t_idx, targ in enumerate(gate_args):
            assert targ >= min_q and targ < max_q
            k = (max_q - min_q) - 1 - (targ - min_q)
            if (1 << k) & i != 0:
                nowi += (1 << (len(gate_args) - 1 - t_idx))

        datas[i] = nowi

    for i in gate_args:
        xor = xor ^ (1 << (max_q - 1 - i))

    for i in range(n):
        nowi = datas[i]
        for j in range(n):
            if (i & xor) == (j & xor):
                nowj = datas[j]
                new_values[i][j] = gate_matrix[nowi][nowj]

    if gpu_output:
        import cupy as cp

        new_values = cp.array(new_values)

    return new_values


def perm_decomposition(permutation: list):
    n = len(permutation)
    assert len(set(permutation)) == n and set(permutation) == set(range(n))

    sorted_idx, swap_pairs = [], []
    idx = 0
    while idx < n:
        if idx in sorted_idx:
            idx += 1
            continue

        target = permutation[idx]
        if target != idx:
            swap_pairs.append([idx, target])
            sorted_idx.append(target)
            permutation[idx], permutation[target] = permutation[target], permutation[idx]
        else:
            idx += 1

    return swap_pairs



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/10 11:09
# @Author  : Han Yu
# @File    : __init__.py

from .gate import *
from .composite_gate import CompositeGate
from .gate_builder import build_gate, build_random_gate, GATE_TYPE_TO_CLASS
from .multicontrol_toffoli import MultiControlToffoli
from .uniformly_control_gate import UniformlyControlGate



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:31 下午
# @Author  : Han Yu
# @File    : instruction_set.py

from QuICT.core.gate import GateType
from .transform_rule.one_qubit_gate_rules import *
from .transform_rule.two_qubit_gate_rules import *


class InstructionSet(object):
    """ InstructionSet describes a set of gates(expectly to be universal set)

    Instruction Set contains gates and some rules, which can be assigned by user.

    Attributes:
        one_qubit_gates(list<GateType>): the types of the one_qubit_gate
        one_qubit_rule(callable): rules to transform SU(2) into instruction set
        two_qubit_gate(GateType): the type of the two_qubit_gate
        rule_map(dictionary): A two-dimensional map from source gate and target gate to transform rule
    """
    # Two-qubit gate and two-qubit rules
    @property
    def two_qubit_gate(self):
        return self.__two_qubit_gate

    @two_qubit_gate.setter
    def two_qubit_gate(self, two_qubit_gate):
        """ set two_qubit_gate

        Args:
            two_qubit_gate(GateType): two-qubit gate in the InstructionSet
        """
        assert isinstance(two_qubit_gate, GateType), TypeError('two_qubit_gate should be a GateType')
        self.__two_qubit_gate = two_qubit_gate

    @property
    def two_qubit_rule_map(self):
        return self.__two_qubit_rule_map

    @two_qubit_rule_map.setter
    def two_qubit_rule_map(self, two_qubit_rule_map):
        self.__two_qubit_rule_map = two_qubit_rule_map

    # One-qubit gates and one-qubit rule
    @property
    def one_qubit_gates(self):
        return self.__one_qubit_gates

    @one_qubit_gates.setter
    def one_qubit_gates(self, one_qubit_gates):
        """ set one_qubit_gates

        Args:
            one_qubit_gates(list<GateType>): one-qubit gates in the InstructionSet
        """
        assert isinstance(one_qubit_gates, list), TypeError('one_qubit_gates should be a list')
        for one_qubit_gate in one_qubit_gates:
            assert isinstance(one_qubit_gate, GateType), TypeError('each one_qubit_gate should be a GateType')
        self.__one_qubit_gates = one_qubit_gates

    @property
    def one_qubit_rule(self):
        """ the rule of decompose 2*2 unitary into target gates

        If not assigned by the register_one_qubit_rule method, some pre-implemented method would be chosen
        corresponding to the one_qubit_gates. An Exception will be raised when no method is chosen.

        Returns:
            callable: the corresponding rule
        """
        if self.__one_qubit_rule:
            return self.__one_qubit_rule
        if set((GateType.rz, GateType.ry)).issubset(set(self.one_qubit_gates)):
            return zyz_rule
        if set((GateType.rz, GateType.rx)).issubset(set(self.one_qubit_gates)):
            return zxz_rule
        if set((GateType.rx, GateType.ry)).issubset(set(self.one_qubit_gates)):
            return xyx_rule
        if set((GateType.h, GateType.rz)).issubset(set(self.one_qubit_gates)):
            return hrz_rule
        if set((GateType.rz, GateType.sx, GateType.x)).issubset(set(self.one_qubit_gates)):
            return ibmq_rule
        if set((GateType.u3)).issubset(set(self.one_qubit_gates)):
            return u3_rule
        raise Exception("please register the SU2 decomposition rule.")

    def __init__(self, two_qubit_gate, one_qubit_gates):
        self.two_qubit_gate = two_qubit_gate
        self.one_qubit_gates = one_qubit_gates
        self.__one_qubit_rule = None
        self.__two_qubit_rule_map = {}

    def select_transform_rule(self, source):
        """ choose a rule which transforms source gate into target gate(2-qubit)

        Args:
            source(GateType): the type of source gate

        Returns:
            callable: the transform rules
        """
        assert isinstance(source, GateType)
        if source in self.two_qubit_rule_map:
            return self.two_qubit_rule_map[source]
        rule = eval(f"{source.name}2{self.two_qubit_gate.name}_rule")
        self.two_qubit_rule_map[source] = rule
        return rule

    def register_one_qubit_rule(self, one_qubit_rule):
        """ register one-qubit gate decompostion rule

        Args:
            one_qubit_rule(callable): decompostion rule
        """
        self.__one_qubit_rule = one_qubit_rule

    def register_two_qubit_rule_map(self, two_qubit_rule, source):
        """ register rule which transforms from source gate into target gate

        Args:
            two_qubit_rule(callable): the transform rule
            source(GateType): the type of source gate
        """
        assert isinstance(source, GateType)
        self.two_qubit_rule_map[source] = two_qubit_rule



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/11 15:36
# @Author  : Xiaoquan Xu
# @File    : rhf_objective.py

import itertools as it

import numpy as np
import scipy.optimize as spopt
import scipy.linalg as splin

from QuICT.chemistry.simulation.parametertensor import ParameterTensor


class RHFObjective:
    """
    The objective for Restricted Hartree-Fock.

    Attributes:
        hamiltonian(ParameterTensor)
        fermion_hamiltonian(FermionOperator)
        num_orbitals(int): number of spacial orbitals
        num_electrons(int)
        nocc(int): number of occupied orbitals
        nvirt(int): number of virtual orbitals
        occ(list): list of positions of occupied orbitals
        virt(list): list of positions of virtual orbitals
    """
    def __init__(self, hamiltonian: ParameterTensor, num_electrons: int):
        self.hamiltonian = hamiltonian
        # In classical simulation it's unnecessary
        # self.fermion_hamiltonian = hamiltonian.get_fermion_operator()
        self.num_orbitals = hamiltonian.obi.shape[0] // 2
        self.num_electrons = num_electrons
        self.nocc = self.num_electrons // 2
        self.nvirt = self.num_orbitals - self.nocc
        self.occ = list(range(self.nocc))
        self.virt = list(range(self.nocc, self.num_orbitals))

    def energy_from_opdm(self, opdm_aa: np.ndarray) -> float:
        """
        According to the opdm(One Particle Density Matrix) parameters,
        generate the tpdm(Two Particles Density Matrix) through Restricted Hartree-Fock conditions
        and calculate the energy by multiplying with the Hamitonian.

        Args:
            opdm_aa: The alpha-alpha block of the RDM

        Returns:
            energy(double)
        """
        opdm = _generate_opdm(opdm_aa)
        tpdm = _generate_tpdm(opdm)
        rdms = ParameterTensor(1, opdm, tpdm)
        return self.hamiltonian.expectation(rdms).real

    def global_gradient_opdm(self, params: np.ndarray, opdm_aa: np.ndarray):
        """
        Generate the gradient

        Args:
            params: The pra
            opdm_aa: The alpha-alpha block of the RDM
        """
        opdm = _generate_opdm(opdm_aa)
        tpdm = _generate_tpdm(opdm)

        # now go through and generate all the necessary Z, Y, Y_kl matrices
        kappa_matrix = rhf_params_to_matrix(params, self.occ, self.virt)
        kappa_matrix_full = np.kron(kappa_matrix, np.eye(2))
        w_full, v_full = np.linalg.eigh(-1j * kappa_matrix_full)  # so that kappa = i U lambda U^
        eigs_scaled_full = get_matrix_of_eigs(w_full)

        grad = np.zeros(self.nocc * self.nvirt, dtype=np.complex128)
        kdelta = np.eye(self.hamiltonian.obi.shape[0])

        # NOW GENERATE ALL TERMS ASSOCIATED WITH THE GRADIENT
        for p in range(self.nocc * self.nvirt):
            grad_params = np.zeros_like(params)
            grad_params[p] = 1
            Y = rhf_params_to_matrix(grad_params, self.occ, self.virt)
            Y_full = np.kron(Y, np.eye(2))

            # Now rotate Y into the basis that diagonalizes Z
            Y_kl_full = v_full.conj().T @ Y_full @ v_full
            # now rotate
            # Y_{kl} * (exp(i(l_{k} - l_{l})) - 1) / (i(l_{k} - l_{l}))
            # into the original basis
            pre_matrix_full = v_full @ (eigs_scaled_full *
                                        Y_kl_full) @ v_full.conj().T

            grad_expectation = -1.0 * np.einsum(
                'ab,pq,aq,pb',
                self.hamiltonian.obi,
                pre_matrix_full,
                kdelta,
                opdm,
                optimize='optimal').real
            grad_expectation += 1.0 * np.einsum(
                'ab,pq,bp,aq',
                self.hamiltonian.obi,
                pre_matrix_full,
                kdelta,
                opdm,
                optimize='optimal').real
            grad_expectation += 1.0 * np.einsum(
                'ijkl,pq,iq,jpkl',
                self.hamiltonian.tbi,
                pre_matrix_full,
                kdelta,
                tpdm,
                optimize='optimal').real
            grad_expectation += -1.0 * np.einsum(
                'ijkl,pq,jq,ipkl',
                self.hamiltonian.tbi,
                pre_matrix_full,
                kdelta,
                tpdm,
                optimize='optimal').real
            grad_expectation += -1.0 * np.einsum(
                'ijkl,pq,kp,ijlq',
                self.hamiltonian.tbi,
                pre_matrix_full,
                kdelta,
                tpdm,
                optimize='optimal').real
            grad_expectation += 1.0 * np.einsum(
                'ijkl,pq,lp,ijkq',
                self.hamiltonian.tbi,
                pre_matrix_full,
                kdelta,
                tpdm,
                optimize='optimal').real
            grad[p] = grad_expectation

        return grad

    def minimization(self, method='CG', verbose=True):
        """
        Perform Hartree-Fock energy minimization.

        Args:
            method: scipy.optimize.minimize method

        Returns:
            Scipy optimize result
        """

        state = []
        for _ in range(self.num_orbitals):
            if _ < self.nocc:
                state += [1]
            else:
                state += [0]
        initial_opdm = np.diag(state)

        def unitary(params):
            kappa = rhf_params_to_matrix(params, self.occ, self.virt)
            return splin.expm(kappa)

        def energy(params):
            u = unitary(params)
            final_opdm_aa = u @ initial_opdm @ np.conjugate(u).T
            return self.energy_from_opdm(final_opdm_aa)

        def gradient(params):
            u = unitary(params)
            final_opdm_aa = u @ initial_opdm @ np.conjugate(u).T
            return self.global_gradient_opdm(params, final_opdm_aa).real

        init_guess = np.zeros(self.nocc * self.nvirt)

        return spopt.minimize(energy,
                              init_guess,
                              jac=gradient,
                              method=method,
                              options={'disp': verbose})


def get_matrix_of_eigs(w: np.ndarray) -> np.ndarray:
    """
    Transform the eigenvalues for getting the gradient.

    Args:
        w: eigenvalues of C-matrix

    Returns:
        New array of transformed eigenvalues
    """
    transform_eigs = np.zeros((w.shape[0], w.shape[0]), dtype=np.complex128)
    for i, j in it.product(range(w.shape[0]), repeat=2):
        if np.isclose(abs(w[i] - w[j]), 0):
            transform_eigs[i, j] = 1
        else:
            transform_eigs[i, j] = (np.exp(1j *
                                           (w[i] - w[j])) - 1) / (1j *
                                                                  (w[i] - w[j]))
    return transform_eigs


def rhf_params_to_matrix(parameters, occ, virt):
    """
    Assemble variational parameters into a anti-Hermitian matrix.

    Args:
        parameters: list of length n_virt * n_occ

    Returns:
        A matrix kappa of size (n_occ + n_virt) * (n_occ + n_virt)

    If i and j both in occ or virt, K[i][j]=0
    If i in virt and j in occ, K[i][j]=parameters[]
    If i in occ and j in virt, K[i][j]=-parameters[]
    """
    # parameters must be real
    for p in parameters:
        if p.imag != 0:
            raise ValueError("parameters input must be real valued")

    kappa = np.zeros((len(occ) + len(virt), len(occ) + len(virt)))
    for idx, (v, o) in enumerate(it.product(virt, occ)):
        kappa[v, o] = parameters[idx].real
        kappa[o, v] = -parameters[idx].real
    return kappa


def _generate_opdm(opdm_aa):
    opdm = np.zeros((opdm_aa.shape[0] * 2, opdm_aa.shape[1] * 2), dtype=np.complex128)
    opdm[::2, ::2] = opdm_aa
    opdm[1::2, 1::2] = opdm_aa
    return opdm


def _generate_tpdm(opdm):
    n = opdm.shape[0]
    tpdm = np.zeros((n, n, n, n), dtype=complex)
    for a in range(n):
        for b in range(n):
            for c in range(n):
                for d in range(n):
                    tpdm[a, b, c, d] = opdm[a, d] * opdm[b, c] - opdm[a, c] * opdm[b, d]\
                        - opdm[b, d] * opdm[a, c] + opdm[b, c] * opdm[a, d]
    return tpdm / 2.



from QuICT.tools.exception import QuICTException


class GateQubitAssignedError(QuICTException):
    """ Gate's qubit assign Error. """
    def __init__(self, msg: str = None):
        super().__init__(1016, msg)


class GateParametersAssignedError(QuICTException):
    """ Gate's Parameter Error. """
    def __init__(self, msg: str = None):
        super().__init__(1017, msg)


class GateAppendError(QuICTException):
    """ BasicGate append error. """
    def __init__(self, msg: str = None):
        super().__init__(1018, msg)


class GateMatrixError(QuICTException):
    """ BasicGate matrix error. """
    def __init__(self, msg: str = None):
        super().__init__(1019, msg)


class CompositeGateAppendError(QuICTException):
    """ CompositeGate append error. """
    def __init__(self, msg: str = None):
        super().__init__(1020, msg)



"""
Decomposition of uniformly-Ry gate, which is specially designed for
the optimization of unitary transform
"""

from typing import *
import numpy as np

from QuICT.core.gate import CompositeGate, CX, CZ, Ry


class UniformlyRyRevision(object):
    def __init__(self, is_cz_left: bool = False):
        self.is_cz_left = is_cz_left

    def execute(self, angle_list):
        """ uniformlyRyGate

        http://cn.arxiv.org/abs/quant-ph/0504100v1 Fig4 a)
        This part is mainly copied from ../uniformly_gate/uniformly_rotation.py
        Here we demand a CZ gate at the edge of the decomposition, therefore the
        recursion process is slightly revised.

        If qubit_num > 2, synthesized gates would have 2 cz gates at rightmost place.
        If qubit_num == 2, there would be only 1 cz gate.

        Args:
            angle_list(list<float>): the angles of Ry Gates

        Returns:
            CompositeGate: the synthesis gate list
        """
        angle_list = list(angle_list)
        n = int(np.round(np.log2(len(angle_list)))) + 1
        if 1 << (n - 1) != len(angle_list):
            raise Exception("the number of parameters unmatched.")
        return self.uniformly_rotation_cz(0, n, angle_list, self.is_cz_left)

    def uniformly_rotation_cz(
        self,
        low: int,
        high: int,
        angles: List[float],
        is_cz_left: bool = False
    ) -> CompositeGate:
        """
        synthesis uniformlyRy gate, bits range [low, high)

        Args:
            low(int): the left range low
            high(int): the right range high
            angles(list<float>): the list of angle y
            mapping(list<int>): the qubit order of gate
            is_cz_left(bool): is cx/cz left decomposition
        Returns:
            gateSet: the synthesis gate list
        """
        return self.inner_uniformly_rotation_cz(low, high, angles, True, is_cz_left)

    def inner_uniformly_rotation_cz(
        self,
        low: int,
        high: int,
        angles: List[float],
        is_first_level: bool,
        is_cz_left: bool = False
    ) -> CompositeGate:
        if low + 1 == high:
            gates = CompositeGate()
            with gates:
                Ry(angles[0].real) & low
            return gates
        length = len(angles) // 2
        Rxp = []
        Rxn = []
        for i in range(length):
            Rxp.append((angles[i] + angles[i + length]) / 2)
            Rxn.append((angles[i] - angles[i + length]) / 2)
        if is_first_level:
            if is_cz_left:
                gates = CompositeGate()
                with gates:
                    CZ & [low, high - 1]
                    # if high - low == 2, no CZ is needed here
                    if high - low > 2:
                        CZ & [low, low + 1]
                gates.extend(self.inner_uniformly_rotation_cz(low + 1, high, Rxn, False, False))
                with gates:
                    # CZ & [low, high - 1]
                    Ry(np.pi / 2) & high - 1
                    CX & [low, high - 1]
                    Ry(-np.pi / 2) & high - 1
                    # The extra CZ derived from CCPhase can also be moved to the edge
                gates.extend(self.inner_uniformly_rotation_cz(low + 1, high, Rxp, False, True))
            else:
                gates = self.inner_uniformly_rotation_cz(low + 1, high, Rxp, False, False)
                with gates:
                    # CZ & [low, high - 1]
                    Ry(np.pi / 2) & high - 1
                    CX & [low, high - 1]
                    Ry(-np.pi / 2) & high - 1
                    # The extra CZ derived from CCPhase can also be moved to the edge
                gates.extend(self.inner_uniformly_rotation_cz(low + 1, high, Rxn, False, True))
                with gates:
                    CZ & [low, high - 1]
                    # if high - low == 2, no CZ is needed here
                    if high - low > 2:
                        CZ & [low, low + 1]
        elif is_cz_left:
            gates = self.inner_uniformly_rotation_cz(low + 1, high, Rxn, False, False)
            with gates:
                CX & [low, high - 1]
            gates.extend(self.inner_uniformly_rotation_cz(low + 1, high, Rxp, False, True))
        else:
            gates = self.inner_uniformly_rotation_cz(low + 1, high, Rxp, False, False)
            with gates:
                CX & [low, high - 1]
            gates.extend(self.inner_uniformly_rotation_cz(low + 1, high, Rxn, False, True))
        return gates



from .gate_type import GateType, MatrixType, SPECIAL_GATE_SET, SUPREMACY_GATE_SET, DIAGONAL_GATE_SET,\
    PAULI_GATE_SET, CLIFFORD_GATE_SET
from .circuit_info import CircuitBased
from .circuit_matrix import CircuitMatrix, get_gates_order_by_depth
from .utils import matrix_product_to_circuit, CGATE_LIST, perm_decomposition
from .id_generator import unique_id_generator



import getpass
import logging
import os
import sys
import socket
from datetime import datetime
from enum import Enum


class LogFormat(Enum):
    """The Enum class of the log format.
    Example:
        - ''LogFormat.zero'': None
        - ''LogFormat.full'': time | host | user | pid | tag | level | msg
        - ''LogFormat.simple'': time | tag | level | msg
        - ''LogFormat.time_only'': time | msg
    """
    zero = None
    full = "%(asctime)s | %(host)s | %(user)s | %(process)d | %(tag)s | %(levelname)s | %(message)s"
    simple = "%(asctime)s | %(tag)s | %(levelname)s | %(message)s"
    time_only = "%(asctime)s | %(message)s"


LEVEL_MAP = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL
}


def msgdecorate(func):
    """The decorator used to construct the log msg."""

    def _decorator(self, msg, *args):
        if not isinstance(msg, str):
            msg = repr(msg)

        if args:
            func(self, "%s %s", msg, repr(args))
        else:
            func(self, "%s", msg)

    return _decorator


class Logger(object):
    """ A wrapper for logging.

    The Logger hosts a file handler and a stdout handler.
    The file handler is set to DEBUG level and will dump all the logging info to the .log
    folder. It is controlled by environment variable "LOG_DUMP".
    The stdout handler's log level is set by the parameter log_level or the environment
    variable "LOG_LEVEL".

    Supported "LOG_LEVEL" includes: DEBUG, INFO, WARN, ERROR, CRITICAL.

    Example:
        $ export LOG_DUMP=True LOG_LEVEL=INFO

    Args:
        tag(str): Log tag for stream and file output.
        format_(LogFormat): Predefined formatter. Defaults to "LogFormat.simple".
        log_level(str): The log level, Defaults to "INFO".
    """

    def __init__(
        self,
        tag: str,
        format_: LogFormat = LogFormat.simple,
        log_level: str = "INFO"
    ):
        self._format = logging.Formatter(fmt=format_.value, datefmt='%Y-%m-%d %H:%M:%S') \
            if format_ is not None else None
        self._stdout_level = os.environ.get("LOG_LEVEL") or LEVEL_MAP[log_level]
        self._logger = logging.getLogger(tag)
        self._logger.setLevel(logging.DEBUG)

        # Set stdout handler
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(self._stdout_level)
        sh.setFormatter(self._format)
        self._logger.addHandler(sh)
        self._extra = {"host": socket.gethostname(), "user": getpass.getuser(), "tag": tag}

        # Set foldout handler
        if os.environ.get("LOG_DUMP", False) == "True":
            log_dump_folder = os.path.join(
                os.path.abspath(os.path.curdir),
                ".log",
                str(os.getpid())
            )
            if not os.path.exists(log_dump_folder):
                os.makedirs(log_dump_folder)

            filename = f"{tag}.{datetime.now().strftime('%Y-%m-%d+%H:%M')}.log"

            # File handler
            fh = logging.FileHandler(filename=f"{os.path.join(log_dump_folder, filename)}", mode='w', encoding="utf-8")
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(self._format)
            self._logger.addHandler(fh)

    @msgdecorate
    def debug(self, msg, *args):
        """Add a log with ''DEBUG'' level."""
        self._logger.debug(msg, *args, extra=self._extra)

    @msgdecorate
    def info(self, msg, *args):
        """Add a log with ''INFO'' level."""
        self._logger.info(msg, *args, extra=self._extra)

    @msgdecorate
    def warn(self, msg, *args):
        """Add a log with ''WARN'' level."""
        self._logger.warning(msg, *args, extra=self._extra)

    @msgdecorate
    def error(self, msg, *args):
        """Add a log with ''ERROR'' level."""
        self._logger.error(msg, *args, extra=self._extra)

    @msgdecorate
    def critical(self, msg, *args):
        """Add a log with ''CRITICAL'' level."""
        self._logger.critical(msg, *args, extra=self._extra)



from QuICT.tools.cli.client import QuICTRemoteManager


remote_manager = QuICTRemoteManager()


def login(username: str, password: str):
    remote_manager.login(username, password)


def logout():
    remote_manager.logout()


def register(username, password, email):
    remote_manager.register(username, password, email)


def unsubscribe(username, password):
    remote_manager.unsubscribe(username, password)


def start_job(file: dict):
    remote_manager.start_job(file)


def delete_job(name: str):
    remote_manager.delete_job(name)


def status_job(name: str):
    remote_manager.status_job(name)


def list_jobs():
    remote_manager.list_jobs()



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/1/11 1:37 下午
# @Author  : Han Yu
# @File    : cnot_store_force.py

import os

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import CX, GateType
from QuICT.qcda.utility import OutputAligner


class CnotStoreForceDepthBfs(object):
    """ use bfs to optimize the cnot circuit

    """
    @OutputAligner()
    def execute(self, circuit: Circuit):
        """ find the best circuit by bfs

        Args:
            input(Circuit): input circuit

        Returns:
            Circuit: optimal circuit

        """
        n = circuit.width()
        if n > 5:
            raise Exception("the qubits number should be smaller than or equal to 5.")
        path = f"{os.path.dirname(os.path.abspath(__file__))}{os.path.sep}utility{os.path.sep}{n}qubit_cnot_depth.inf"
        if not os.path.exists(path):
            from .utility.cnot_depth_bfs import generate_json
            generate_json(n)
        with open(path, "r") as f:
            loadnow = f.readline()

        circuit_opt = Circuit(n)
        input_matrix = np.identity(n, dtype=bool)
        goal = 0
        for gate in circuit.gates:
            if gate.type != GateType.cx:
                raise Exception("the circuit should only contain CX gate")
            input_matrix[gate.targ, :] = input_matrix[gate.targ, :] ^ input_matrix[gate.carg, :]
        for i in range(n):
            for j in range(n):
                if input_matrix[i, j]:
                    goal ^= 1 << (i * n + j)

        goal_string = f";{goal}:"
        index = loadnow.find(goal_string)
        if index == -1:
            raise Exception("generate error")
        begin = index + len(goal_string)
        end = loadnow.find(";", begin)
        tuples_encode = loadnow[begin:end].split(",")
        ans = []
        for tuple_encode in tuples_encode:
            if len(tuple_encode) > 0:
                ans.append([int(tuple_encode) // 5, int(tuple_encode) % 5])
        for plist in ans:
            CX | circuit_opt(plist)
        return circuit_opt



import cupy as cp
import numpy as np

from .gate_function import prop_add_double_kernel, prop_add_single_kernel, MeasureGate_prop


__outward_functions = [
    "complex_multiply",
    "float_multiply"
]


complex_multiply_single = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void SimpleMultiply(complex<float> val, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        vec[label] = vec[label]*val;
    }
    ''', 'SimpleMultiply')
complex_multiply_single.compile()


complex_multiply_double = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void SimpleMultiply(complex<double> val, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        vec[label] = vec[label]*val;
    }
    ''', 'SimpleMultiply')
complex_multiply_double.compile()


float_multiply_single = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void FloatMultiply(const float value, complex<float>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        vec[label] = vec[label] * value;
    }
    ''', 'FloatMultiply')
float_multiply_single.compile()


float_multiply_double = cp.RawKernel(r'''
    #include <cupy/complex.cuh>
    extern "C" __global__
    void FloatMultiply(const double value, complex<double>* vec) {
        int label = blockDim.x * blockIdx.x + threadIdx.x;

        vec[label] = vec[label] * value;
    }
    ''', 'FloatMultiply')
float_multiply_double.compile()


def complex_multiply(val, vec, vec_bit, sync: bool = False):
    task_number = 1 << vec_bit
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        complex_multiply_single(
            (block_num,),
            (thread_per_block,),
            (val, vec)
        )
    else:
        complex_multiply_double(
            (block_num,),
            (thread_per_block,),
            (val, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()


def float_multiply(val, vec, vec_bit, sync: bool = False):
    task_number = 1 << vec_bit
    thread_per_block = min(256, task_number)
    block_num = task_number // thread_per_block

    if vec.dtype == np.complex64:
        float_multiply_single(
            (block_num,),
            (thread_per_block,),
            (val, vec)
        )
    else:
        float_multiply_double(
            (block_num,),
            (thread_per_block,),
            (val, vec)
        )

    if sync:
        cp.cuda.Device().synchronize()



from typing import DefaultDict
import numpy as np

from QuICT.core.layout import Layout


class SupremacyLayout(Layout):
    def __init__(self, qubits: int, name: str = "unknown"):
        assert(qubits >= 5)
        Layout.__init__(self, qubits, name)

        self._build_supremacy_layout()
        self._add_supremacy_edge()

    def _build_supremacy_layout(self):
        if self.qubit_number == 5:
            self.qubit_table = np.array(
                [[0, 1],
                 [2, -1],
                 [3, 4]], dtype=np.int32
            )
        else:
            full_size_width = int(np.ceil(np.sqrt(self.qubit_number)))

            self.qubit_table = np.zeros((full_size_width, full_size_width), dtype=np.int32)
            self.qubit_table -= 1

            for i in range(self.qubit_number):
                row = i // (full_size_width - 1)
                col = i % (full_size_width - 1)

                if row >= full_size_width:
                    self.qubit_table[i % full_size_width, full_size_width - 1] = i
                else:
                    self.qubit_table[row, col] = i

    def _add_supremacy_edge(self):
        row, col = self.qubit_table.shape
        self._pattern_edge = DefaultDict(list)

        for i in range(1, row, 2):
            for j in range(col):
                current_qubit = self.qubit_table[i, j]
                if current_qubit != -1:
                    # pattern "A"
                    if i + 1 < row:
                        qubit_a = self.qubit_table[i + 1, j]
                        if qubit_a != -1:
                            self._pattern_edge["A"].append([current_qubit, qubit_a])
                            self.add_edge(current_qubit, qubit_a)

                    # pattern "B"
                    if j + 1 < col:
                        qubit_b = self.qubit_table[i - 1, j + 1]
                        if qubit_b != -1:
                            self._pattern_edge["B"].append([current_qubit, qubit_b])
                            self.add_edge(current_qubit, qubit_b)

                    # pattern "C"
                    qubit_c = self.qubit_table[i - 1, j]
                    if qubit_c != -1:
                        self._pattern_edge["C"].append([current_qubit, qubit_c])
                        self.add_edge(current_qubit, qubit_c)

                    # pattern "D"
                    if i + 1 < row and j + 1 < col:
                        qubit_d = self.qubit_table[i + 1, j + 1]
                        if qubit_d != -1:
                            self._pattern_edge["D"].append([current_qubit, qubit_d])
                            self.add_edge(current_qubit, qubit_d)

    def get_edges_by_pattern(self, pattern: str):
        return self._pattern_edge[pattern]



import os
import shutil


# Default User Path
USER_ROOT_PATH = "/home/likaiqi/Cluster/User"


def create_user_folder(user_name):
    user_path = os.path.join(USER_ROOT_PATH, user_name)
    if not os.path.exists(user_path):
        os.makedirs(user_path)


def create_job_folder(user_name, job_name):
    job_path = os.path.join(USER_ROOT_PATH, user_name, job_name)
    if not os.path.exists(job_path):
        os.makedirs(job_path)
    else:
        raise KeyError("job name already exists.")

    return job_path


def delete_user_folder(user_name):
    user_path = os.path.join(USER_ROOT_PATH, user_name)
    if os.path.exists(user_path):
        shutil.rmtree(user_path)


def delete_job_folder(user_name, job_name):
    job_path = os.path.join(USER_ROOT_PATH, user_name, job_name)
    if os.path.exists(job_path):
        shutil.rmtree(job_path)



from typing import *
import numpy as np

from QuICT.core import *
from QuICT.core.gate import *


def add_factor_shift_into_phase(gates: CompositeGate, shift: complex) -> CompositeGate:
    phase = np.angle(shift)
    phase_gate = GPhase(phase) & 0
    gates.append(phase_gate)
    return gates


def quantum_shannon_decompose(
        u1: np.ndarray,
        u2: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Decompose a block diagonal even-size unitary matrix.
    block_diag(u1,u2) == block_diag(v, v) @ block_diag(d, d_dagger) @ block_diag(w, w)

    Args:
        u1 (np.ndarray): upper-left block
        u2 (np.ndarray): right-bottom block

    Returns:
        Tuple[np.ndarray,np.ndarray,np.ndarray]
    """
    s = u1 @ u2.conj().T

    eig_values, v = np.linalg.eig(s)
    v_dagger = v.conj().T
    d = np.sqrt(np.diag(eig_values))
    w = d @ v_dagger @ u2

    return v, d, w


def shift_ratio(mat1: np.ndarray, mat2: np.ndarray) -> complex:
    return mat1.dot(np.linalg.inv(mat2))[0, 0]



import os
import os.path as osp
import re
from typing import Dict, List, Tuple, Union

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter

from QuICT.core import *
from QuICT.core.gate.composite_gate import CompositeGate
from QuICT.qcda.mapping.ai.data_def import CircuitInfo, ReplayMemory, State, StateSlim, TrainConfig, ValidationData
from QuICT.qcda.mapping.ai.net.nn_mapping import NnMapping
from QuICT.qcda.mapping.ai.net.rl_agent import Agent
from QuICT.tools.interface import OPENQASMInterface
from QuICT.tools.logger import Logger

logger = Logger("rl-mapping-learner")


def _wrap2circ(cg_or_circ: Union[Circuit, CompositeGate] = None):
    if cg_or_circ is None:
        return None
    elif isinstance(cg_or_circ, Circuit):
        return cg_or_circ
    elif isinstance(cg_or_circ, CompositeGate):
        circ = Circuit(cg_or_circ.width())
        circ.extend(cg_or_circ.gates)
        return circ
    else:
        raise TypeError("Only supports Circuit/CompositeGate")


class Learner:
    def __init__(self, config: TrainConfig) -> None:
        self.config = config

        if config.inference:
            logger.info("Preparing policy network...")
            try:
                model_path = config.inference_model_dir
                model_path = osp.join(model_path, f"{config.topo.name}.pt")
                self._policy_net = NnMapping(
                    qubit_num=config.topo.qubit_number,
                    max_gate_num=config.max_gate_num,
                    feat_dim=config.feat_dim,
                    action_num=config.action_num,
                    device=config.device,
                )
                states = torch.load(model_path)
                self._policy_net.load_state_dict(states)
                self._policy_net = self._policy_net.to(device=config.device)
                self._policy_net.train(not config.inference)
            except:
                raise Exception("Cannot load model for " + model_path)
        elif not config.inference:
            logger.info("Preparing policy network...")
            self._policy_net = NnMapping(
                qubit_num=config.topo.qubit_number,
                max_gate_num=config.max_gate_num,
                feat_dim=config.feat_dim,
                action_num=config.action_num,
                device=config.device,
            ).to(device=config.device)
            self._policy_net.train(not config.inference)
            logger.info("Preparing target network...")
            self._target_net = NnMapping(
                qubit_num=config.topo.qubit_number,
                max_gate_num=config.max_gate_num,
                feat_dim=config.feat_dim,
                action_num=config.action_num,
                device=config.device,
            ).to(device=config.device)
            self._target_net.train(config.inference)
            # Guarantee they have the same parameter values.
            self._target_net.load_state_dict(self._policy_net.state_dict())

            logger.info("Preparing experience pool...")
            self.replay = ReplayMemory(config.replay_pool_size)

            # Loss function & optimizer
            self._smooth_l1 = nn.SmoothL1Loss()
            self._optimizer = optim.RMSprop(
                self._policy_net.parameters(),
                lr=self.config.lr,
            )

            # Prepare path to save model files during training
            logger.info("Preparing model saving directory...")
            if not osp.exists(config.model_path):
                os.makedirs(config.model_path)
            self._model_path = config.model_path

            # Prepare summary writer and its logging directory
            if not osp.exists(config.log_dir):
                os.makedirs(config.log_dir)
            self._writer = SummaryWriter(log_dir=config.log_dir)

            # Validation circuits
            logger.info("Preparing validation data...")
            self._v_data: List[ValidationData] = []
            self._mcts_num: Dict[str, int] = {}
            self._load_v_data()

            # best so far
            self._v_best_score = 1e9

    def _load_v_data(self):
        v_data_dir = osp.join(self.config.factory._data_dir, "v_data")
        pattern = re.compile(r"(\w+)_(\d+).qasm")
        for _, _, file_names in os.walk(v_data_dir):
            for file_name in file_names:
                if file_name.startswith("mapped"):
                    continue

                file_path = osp.join(v_data_dir, file_name)
                # mapped_file_path = osp.join(v_data_dir, f"mapped_{file_name}")
                groups = pattern.match(file_name)
                topo_name = groups[1]
                if topo_name != self.config.topo.name:
                    continue
                topo = self.config.factory.topo_map[topo_name]
                circ = OPENQASMInterface.load_file(file_path).circuit
                # mcts_mapped_circ = OPENQASMInterface.load_file(mapped_file_path).circuit
                self._v_data.append(
                    ValidationData(
                        circ=circ,
                        topo=topo,
                    )
                )
        self._v_data.sort(key=lambda x: len(x.circ.gates))

    def map_all(
        self,
        circ: Union[Circuit, CompositeGate],
        layout: Layout,
        policy_net: NnMapping,
        policy_net_device: str,
        cutoff: int = None,
    ) -> Tuple[CompositeGate, CompositeGate]:
        """Map given circuit to topology layout. Note that this methods will change internal exploration state.

        Args:
            circ (CircuitBased): Circuit to be mapped.
            layout (Layout): Topology layout.
            policy_net (GnnMapping): Policy network.
            policy_net_device (str): Policy network device.
            cutoff (int, optional): The maximal steps of execution. If the policy network cannot
                stop with in `cutoff` steps, force it to early stop. Defaults to None.

        Returns:
            Tuple[CompositeGate, CompositeGate]: Mapped circuit, remained circuit.
        """
        logic2phy = [i for i in range(layout.qubit_number)]
        agent = Agent(config=self.config)
        topo_graph = self.config.factory.get_topo_graph(topo=layout)
        circ_state = CircuitInfo(circ=circ, max_gate_num=policy_net._max_gate_num)

        circ_state.eager_exec(
            logic2phy=logic2phy, topo_graph=topo_graph, physical_circ=agent.mapped_circ
        )
        agent.state = State(
            circ_info=circ_state,
            layout=layout,
            logic2phy=logic2phy,
        )

        terminated = circ_state.count_gate() == 0
        step = 0
        with torch.no_grad():
            while not terminated:
                action = agent.select_action(
                    policy_net=policy_net,
                    policy_net_device=policy_net_device,
                )
                _, _, _, terminated = agent.take_action(action=action)
                step += 1
                if cutoff is not None and step >= cutoff:
                    break
            return agent.mapped_circ, agent.state.remained_circ()

    def validate_model(self) -> List[ValidationData]:
        """Map all validation circuits to corresponding topologies.

        Returns:
            List[ValidationData]: Multiple instances.
        """
        for v_datum in self._v_data:
            cutoff = len(v_datum.circ.gates) * v_datum.circ.width()
            result_circ, remained_circ = self.map_all(
                max_gate_num=self.config.max_gate_num,
                circ=v_datum.circ,
                layout=v_datum.topo,
                policy_net=self._target_net,
                policy_net_device=self.config.device,
                cutoff=cutoff,
            )
            v_datum.rl_mapped_circ = _wrap2circ(result_circ)
            v_datum.remained_circ = _wrap2circ(remained_circ)
        return self._v_data

    def optimize_model(self) -> Union[None, float]:
        if len(self.replay) < self.config.batch_size:
            print(
                f"Experience pool is too small({len(self.replay):2}). Keep exploring..."
            )
            return None
        transitions = self.replay.sample(self.config.batch_size)
        states, actions, next_states, rewards = zip(*transitions)

        actions = torch.tensor(
            [[self.config.action_id_by_swap[action]] for action in actions],
            dtype=torch.long,
            device=self.config.device,
        )  # [B, 1]

        rewards = torch.tensor(rewards, device=self.config.device)

        data_list = [state.to_nn_data() for state in states]
        data_batch = StateSlim.batch_from_list(
            data_list=data_list, device=self.config.device
        )

        # Current Q estimation
        q_vec = self._policy_net(
            data_batch.x,
            data_batch.edge_index,
            data_batch.batch,
        )  # [b, a_gn + a_gs]
        state_action_values = q_vec.gather(1, actions).squeeze()

        # Q* by Bellman Equation
        non_final_mask = torch.tensor(
            tuple(map(lambda s: s is not None, next_states)),
            device=self.config.device,
            dtype=torch.bool,
        )
        non_final_data_list = [
            state.to_nn_data() for state in next_states if state is not None
        ]
        non_final_data_batch = StateSlim.batch_from_list(
            data_list=non_final_data_list, device=self.config.device
        )

        #
        # Use normal DQN
        #
        next_state_values = torch.zeros(
            self.config.batch_size, device=self.config.device
        )
        next_state_values[non_final_mask] = (
            self._target_net(
                non_final_data_batch.x,
                non_final_data_batch.edge_index,
                non_final_data_batch.batch,
            )  # [b, a]
            .clone()
            .detach()
            .max(1)[0]
        )

        expected_state_action_values = (next_state_values * self.config.gamma) + rewards

        loss = self._smooth_l1(state_action_values, expected_state_action_values)

        self._optimizer.zero_grad()
        loss.backward()
        loss_val = loss.item()
        self._optimizer.step()
        return loss_val

    def show_validation_results(self, results: List[ValidationData], epoch_id: int):
        logger.info("[Validation Summary]")
        rl_num = {}
        original_num = {}
        for v_datum in results:
            topo_name = v_datum.topo.name
            rl_num[topo_name] = 0
            original_num[topo_name] = 0
        for idx, v_datum in enumerate(results):
            topo_name = v_datum.topo.name
            _rl_num = len(v_datum.rl_mapped_circ.gates)
            _original_num = len(v_datum.circ.gates)
            rl_num[topo_name] += _rl_num
            original_num[topo_name] += _original_num
            print(
                f"    #Gate in circuit: {_original_num}, #Gate by RL: {_rl_num} ({topo_name})"
            )
        if rl_num[topo_name] < self._v_best_score:
            self._v_best_score = rl_num[topo_name]
            if epoch_id > 5:
                torch.save(
                    self._target_net.state_dict(),
                    osp.join(self._model_path, f"model_epoch_{epoch_id}.pt"),
                )
        self._writer.add_scalars(
            f"Validation Performance ({topo_name})",
            {
                "#Gate in circuit": original_num[topo_name],
                "#Gate by RL": rl_num[topo_name],
                "#Gate by RL (best)": self._v_best_score,
            },
            epoch_id + 1,
        )
        logger.info()



import heapq
from collections import deque
from typing import List, Tuple

from .matching_dag_circuit import (Match, MatchingDAGCircuit, MatchingDAGNode,
                                   NodeInfo)


class MatchingScenario:
    """
    Class for matching scenario in backward searching tree
    """
    def __init__(self, circuit_info, template_info, match, counter):
        """
        Args:
            circuit_info(List[NodeInfo]): node infos of circuit
            template_info(List[NodeInfo]): node infos of template
            match(List[Tuple[int, int]]): current match info
            counter(int): number of gates visited
        """
        self.circuit_info = circuit_info
        self.template_info = template_info
        self.match = match.copy()
        self.counter = counter


class BackwardMatch:
    """
    Class for backward matching.
    """

    @classmethod
    def _calc_gate_indices(cls, circuit: MatchingDAGCircuit):
        """
        Returns:
             List[int]: the list of gate to backward match
        """
        return list(filter(
            lambda i: circuit.get_node(i).matchable(),
            reversed(range(circuit.size))
        ))

    @classmethod
    def _find_candidates(cls, match, template, template_info, t_node_id) -> List[MatchingDAGNode]:
        """
        Find candidate for next match.
        """
        successors = set(template.all_successors(t_node_id))

        matches = {m[0] for m in match}
        ret = sorted(filter(
            lambda x: x not in successors and x not in matches and not template_info[x].is_blocked,
            range(t_node_id + 1, template.size)
        ), reverse=True)

        return [template.get_node(i) for i in ret]

    @classmethod
    def _prune(cls, scenarios: deque, gate_indices, depth, width):
        """
        Heuristic backward matching algorithm.
        """
        counters = [s.counter for s in scenarios]
        if counters.count(max(counters)) == len(counters) and counters[0] < len(gate_indices) \
                and counters[0] % depth == 0:
            ret = deque(heapq.nlargest(width, scenarios, key=lambda s: len(s.match)))
            return ret
        return scenarios

    @classmethod
    def _left_block(cls,
                    circuit: MatchingDAGCircuit,
                    template: MatchingDAGCircuit,
                    scenario: MatchingScenario,
                    c_node_id: int):
        """
        Block a node and all its predecessors in circuit.
        """
        c_info = scenario.circuit_info.copy()
        t_info = scenario.template_info.copy()
        match = scenario.match.copy()
        counter = scenario.counter

        c_info[c_node_id] = NodeInfo(None, True)
        for c_nxt_id in circuit.all_predecessors(c_node_id):
            c_info[c_nxt_id] = NodeInfo(None, True)

        nxt_scenario = MatchingScenario(c_info, t_info, match, counter + 1)
        return nxt_scenario

    @classmethod
    def _right_block(cls,
                     circuit: MatchingDAGCircuit,
                     template: MatchingDAGCircuit,
                     scenario: MatchingScenario,
                     c_node_id: int):
        """
        Block a node and all its successors in circuit.
        """
        c_info = scenario.circuit_info.copy()
        t_info = scenario.template_info.copy()
        match = scenario.match
        counter = scenario.counter

        c_info[c_node_id] = NodeInfo(None, True)

        broken_match = []
        for c_nxt_id in circuit.all_successors(c_node_id):
            t_nxt_id = c_info[c_nxt_id].matched_with
            if t_nxt_id is not None:
                broken_match.append(t_nxt_id)
                t_info[t_nxt_id] = NodeInfo(None, False)
            c_info[c_nxt_id] = NodeInfo(None, True)

        new_match = list(filter(lambda x: x[0] not in broken_match, match))
        nxt_scenario = MatchingScenario(c_info, t_info, new_match, counter + 1)

        return broken_match, nxt_scenario

    @classmethod
    def execute(cls,
                circuit: MatchingDAGCircuit,
                template: MatchingDAGCircuit,
                forward_match: List[Tuple[int, int]],
                c_node_id: int,
                t_node_id: int,
                qubit_mapping: List[int],
                prune_param=None) -> List[Match]:
        """
        Execute backward matching algorithm.

        Args:
            circuit(MatchingDAGCircuit): the circuit to match
            template(MatchingDAGCircuit): the template to be matched
            forward_match(List[Tuple[int, int]]): matches by Forward matching
            c_node_id(int): the starting node of `circuit`
            t_node_id(int): the starting node of `template`
            qubit_mapping(List[int]): qubit mapping
                (qubit i in `temlate` is mapped to qubit_mapping[i] is circuit)
            prune_param(List[int]): heuristic backward matching parameters

        Returns:
            List[Match]: List of maximal matches found
        """

        # the list of id's of gate to match in circuit
        gate_indices = cls._calc_gate_indices(circuit)
        scenarios = deque([
            MatchingScenario(
                circuit.matching_info(),
                template.matching_info(),
                forward_match,
                0
            )
        ])

        # remaining number of gates to match in template (given by the paper, not sure why)
        remain_cnt = template.size - t_node_id - len(forward_match)

        res = []
        while len(scenarios) > 0:
            if prune_param is not None:
                scenarios = cls._prune(scenarios, gate_indices, *prune_param)

            # extract current scenario
            cur_scenario = scenarios.popleft()
            c_info = cur_scenario.circuit_info
            t_info = cur_scenario.template_info
            match = cur_scenario.match
            counter = cur_scenario.counter
            backward_match = list(filter(lambda x: x not in forward_match, match))

            # match found if all nodes in circuit visited or all gates in template matched
            if counter >= len(gate_indices) or len(backward_match) == remain_cnt:
                res.append(Match(match, qubit_mapping))
                continue

            cur_c_node_id = gate_indices[counter]
            cur_c_node: MatchingDAGNode = circuit.get_node(cur_c_node_id)
            if c_info[cur_c_node_id].is_blocked:
                # continue if current node cannot be matched
                nxt_scenario = MatchingScenario(c_info, t_info, match, counter + 1)
                scenarios.append(nxt_scenario)
                continue

            cands = cls._find_candidates(match, template, t_info, t_node_id)
            flag_broken = True  # if all possible matches will break sth
            flag_succeed = False  # if there is possible match

            for cur_t_node in cands:
                # FIXME remove identical matches

                if cur_t_node.compare_with(cur_c_node, qubit_mapping):
                    # option 1.1: match the gate and block unconnected matches
                    cur_c_info = c_info.copy()
                    cur_t_info = t_info.copy()
                    cur_match = match.copy()

                    broken_match = []

                    for block_id in template.all_successors(cur_t_node.id):
                        if cur_t_info[block_id].matched_with is None:
                            cur_t_info[block_id] = NodeInfo(None, True)

                            for t_nxt_id in template.all_successors(block_id):
                                c_nxt_id = cur_t_info[t_nxt_id].matched_with
                                if c_nxt_id is not None:
                                    cur_c_info[c_nxt_id] = NodeInfo(None, True)
                                    broken_match.append(t_nxt_id)
                                cur_t_info[t_nxt_id] = NodeInfo(None, True)

                    flag_broken &= len(broken_match) > 0
                    new_match = list(filter(lambda x: x[0] not in broken_match, cur_match))

                    # check if fixed match is unchanged
                    if (t_node_id, c_node_id) in new_match and \
                            all(map(lambda x: x in new_match, backward_match)):
                        cur_t_info[cur_t_node.id] = NodeInfo(cur_c_node_id, False)
                        cur_c_info[cur_c_node_id] = NodeInfo(cur_t_node.id, False)
                        new_match.append((cur_t_node.id, cur_c_node_id))
                        nxt_scenario = MatchingScenario(cur_c_info, cur_t_info, new_match, counter + 1)
                        scenarios.append(nxt_scenario)

                        flag_succeed = True

            if flag_succeed:
                # option 1.2: right block the node (available if matching succeeded)
                broken_match, nxt_scenario = cls._right_block(circuit, template, cur_scenario, cur_c_node_id)

                if (t_node_id, c_node_id) in nxt_scenario.match and \
                        all(map(lambda x: x in nxt_scenario.match, backward_match)):
                    scenarios.append(nxt_scenario)

                if broken_match and flag_broken:
                    # option 1.3: left block the node (available if all above options break previous matches)
                    nxt_scenario = cls._left_block(circuit, template, cur_scenario, cur_c_node_id)
                    scenarios.append(nxt_scenario)
            else:
                # option 2: if no match found, block it
                c_info[cur_c_node_id] = NodeInfo(None, True)
                following_match = list(filter(
                    lambda i: c_info[i].matched_with is not None,
                    circuit.all_successors(cur_c_node_id)
                ))

                if not following_match or not cur_c_node.predecessors:
                    # option 2.1: if block it affect nothing, do it.
                    nxt_scenario = MatchingScenario(c_info, t_info, match, counter + 1)
                    scenarios.append(nxt_scenario)
                else:
                    # option 2.2: left block it
                    nxt_scenario = cls._left_block(circuit, template, cur_scenario, cur_c_node_id)
                    scenarios.append(nxt_scenario)

                    # option 2.3: right block it
                    broken_match, nxt_scenario = cls._right_block(circuit, template, cur_scenario, cur_c_node_id)
                    if (t_node_id, c_node_id) in nxt_scenario.match and \
                            all(map(lambda x: x in nxt_scenario.match, backward_match)):
                        scenarios.append(nxt_scenario)

        # return distinct matches of maximal length
        max_len = max(len(match) for match in res)
        maximal_res = filter(lambda m: len(m) == max_len, res)
        return list(set(maximal_res))



import random
import numpy as np

from numba.pycc import CC
from numba.types import bool_

from QuICT.ops.utils import mapping_augment


cc = CC("linalg")


@cc.export('matrixtensorf', 'c8[:, :](c8[:, :], i4, i4)')
@cc.export('matrixtensord', 'c16[:, :](c16[:, :], i4, i4)')
def MatrixTensorI(A, n, m):
    """ tensor I^n and A and I^m

    Args:
        A(np.array<np.complex>): the matrix A
        n(int): the index of indentity
        m(int): the index of indentity

    Returns:
        np.array<np.complex>: the tensor result I^n ⊗ A ⊗ I^m
    """
    i_m = np.identity(m)
    row_a, col_a = A.shape
    MatrixTensor = np.zeros((n * m * row_a, n * m * col_a), dtype=A.dtype)

    for i in range(row_a):
        for j in range(col_a):
            temp_M = A[i, j] * i_m
            for k in range(n):
                start_row_idx = k * m * row_a + i * m
                start_col_idx = k * m * col_a + j * m
                MatrixTensor[start_row_idx:start_row_idx + m, start_col_idx:start_col_idx + m] = temp_M

    return MatrixTensor


@cc.export('matrixpermf', 'c8[:, :](c8[:, :], i4[:], bool_)')
@cc.export('matrixpermd', 'c16[:, :](c16[:, :], i4[:], bool_)')
def MatrixPermutation(A: np.ndarray, mapping: np.ndarray, changeInput: bool = False) -> np.ndarray:
    """ permute A with mapping, inplace

    Args:
        A(np.array<np.complex>): the matrix A
        mapping(np.ndarray<int>): the qubit mapping
        changeInput(bool): whether changes in A

    """
    if not A.shape[0] == (1 << mapping.shape[0]):
        raise IndexError("Indices do not match!")

    idx_mapping = mapping_augment(mapping)

    # Do NOT perform parallel operations over row permutations!
    # They are just too spare in memory. Elements in the same column
    # are distributed with a gap as matrix row length.
    perm_mat = A[idx_mapping, :]
    for i in range(idx_mapping.shape[0]):
        perm_mat[i] = perm_mat[i][idx_mapping]

    if changeInput:
        A[:, :] = perm_mat

    return perm_mat


@cc.export('vectorpermf', 'c8[:](c8[:], i4[:], bool_)')
@cc.export('vectorpermd', 'c16[:](c16[:], i4[:], bool_)')
def VectorPermutation(A: np.ndarray, mapping: np.ndarray, changeInput: bool = False):
    """ permutaion A with mapping, changeInput

    Args:
        A(np.array<np.complex>): the matrix A
        mapping(np.ndarray<int>): the qubit mapping
        changeInput(bool): whether changes in A
    Returns:
        np.array<np.complex>: the result of Permutation
    """
    if not A.shape[0] == 1 << mapping.shape[0]:
        raise IndexError("Indices do not match!")

    switched_idx = mapping_augment(mapping)

    if changeInput:
        A[:] = A[switched_idx]

    return A[switched_idx]


@cc.export('tensorf', 'c8[:, :](c8[:, :], c8[:, :])')
@cc.export('tensord', 'c16[:, :](c16[:, :], c16[:, :])')
def tensor(A, B):
    """ tensor A and B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B

    Returns:
        np.array<np.complex>: the tensor result A ⊗ B
    """
    row_a, col_a = A.shape
    row_b, col_b = B.shape
    tensor_data = np.empty((row_a * row_b, col_a * col_b), dtype=A.dtype)

    for r in range(row_a):
        for c in range(col_a):
            tensor_data[r * row_b:(r + 1) * row_b, c * col_b:(c + 1) * col_b] = A[r, c] * B

    return tensor_data


@cc.export('dotf', 'c8[:, :](c8[:, :], c8[:, :])')
@cc.export('dotd', 'c16[:, :](c16[:, :], c16[:, :])')
@cc.export('vdotmf', 'c8[:](c8[:, :], c8[:])')
@cc.export('vdotmd', 'c16[:](c16[:, :], c16[:])')
def dot(A, B):
    """ dot matrix A and matrix B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B

    Returns:
        np.array<np.complex>: A * B
    """
    return np.dot(A, B)


@cc.export('multiplyf', 'c8[:, :](c8[:, :], c8[:, :])')
@cc.export('multiplyd', 'c16[:, :](c16[:, :], c16[:, :])')
def multiply(A, B):
    return np.multiply(A, B)


@cc.export('matdotvecf', 'c8[:](c8[:, :], i4, c8[:], i4, i4[:])')
@cc.export('matdotvecd', 'c16[:](c16[:, :], i4, c16[:], i4, i4[:])')
def matrix_dot_vector(
    mat,
    mat_bit,
    vec,
    vec_bit,
    affect_args
):
    repeat = 1 << (vec_bit - mat_bit)
    arg_len = 1 << mat_bit
    sorted_args = affect_args.copy()
    sorted_args.sort()
    aux = np.zeros_like(vec)
    for i in range(repeat):
        for sarg_idx in range(mat_bit):
            less = i & ((1 << sorted_args[sarg_idx]) - 1)
            i = i >> sorted_args[sarg_idx] << (sorted_args[sarg_idx] + 1) + less

        indexes = np.array([i] * arg_len, dtype=np.int32)
        for i in range(1, arg_len, 1):
            for j in range(mat_bit):
                if i & (1 << j):
                    indexes[i] += 1 << affect_args[j]

        aux[indexes] = np.dot(mat, vec[indexes])

    return aux


@cc.export('measuref', 'bool_(i4, c8[:])')
@cc.export('measured', 'bool_(i4, c16[:])')
def measure_gate_apply(
    index: int,
    vec: np.array
):
    target_index = 1 << index
    vec_idx_0 = [idx for idx in range(len(vec)) if not idx & target_index]
    vec_idx_0 = np.array(vec_idx_0, dtype=np.int32)
    vec_idx_1 = [idx for idx in range(len(vec)) if idx & target_index]
    vec_idx_1 = np.array(vec_idx_1, dtype=np.int32)
    prob = np.sum(np.square(np.abs(vec[vec_idx_0])))

    _1 = random.random() > prob
    if _1:
        alpha = np.float64(1 / np.sqrt(1 - prob))
        vec[vec_idx_0] = np.complex128(0)
        vec[vec_idx_1] = vec[vec_idx_1] * alpha
    else:
        alpha = np.float64(1 / np.sqrt(prob))
        vec[vec_idx_0] = vec[vec_idx_0] * alpha
        vec[vec_idx_1] = np.complex128(0)

    return _1


if __name__ == "__main__":
    cc.compile()



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/22 7:14
# @Author  : Han Yu
# @File    : __init__.py

from .cnot_force import CnotForceBfs
from .cnot_force_depth import CnotForceDepthBfs
from .cnot_store_force import CnotStoreForceBfs
from .cnot_store_force_depth import CnotStoreForceDepthBfs
from .cnot_local_force import CnotLocalForceBfs
from .cnot_local_force_depth import CnotLocalForceDepthBfs



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/22 7:14
# @Author  : Han Yu
# @File    : cnot_force.py

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import CX, GateType
from QuICT.qcda.utility import OutputAligner
from QuICT.qcda.optimization.cnot_local_force.utility.utility import apply_cx, path


class CnotForceBfs(object):
    """
    use bfs to optimize the cnot circuit
    """
    @OutputAligner()
    def execute(self, circuit: Circuit):
        """ find the best circuit by bfs

        Args:
            input(Circuit): input circuit

        Returns:
            Circuit: optimal circuit

        """
        n = circuit.width()
        circuit_opt = Circuit(n)
        input_matrix = np.identity(n, dtype=bool)
        now = 0
        goal = 0
        for gate in circuit.gates:
            if gate.type != GateType.cx:
                raise Exception("the circuit should only contain CX gate")
            input_matrix[gate.targ, :] = input_matrix[gate.targ, :] ^ input_matrix[gate.carg, :]
        for i in range(n):
            for j in range(n):
                if input_matrix[i, j]:
                    goal ^= 1 << (i * n + j)
        for i in range(n):
            now ^= (1 << (i * n + i))
        vis = np.zeros(1 << (n * n), dtype=int)
        vis[now] = 1
        pre = [None] * (1 << (n * n))
        queue = [now]
        l = 0
        if now == goal:
            return circuit_opt

        while True:
            now = queue[l]
            for i in range(n):
                for j in range(n):
                    if i != j:
                        new_state = apply_cx(now, i, j, n)
                        if vis[new_state] == 0:
                            pre[new_state] = path(now, i, j)
                            vis[new_state] = vis[now] + 1
                            if new_state == goal:
                                paths = []
                                while new_state != queue[0]:
                                    paths.append(pre[new_state].CX_tuple)
                                    new_state = pre[new_state].father_node
                                for index in range(len(paths) - 1, -1, -1):
                                    CX | circuit_opt(list(paths[index]))

                                return circuit_opt
                            queue.append(new_state)
            l += 1
